__all__ = (
    "DitherDetailer",
    "EuclidDitherDetailer",
    "CameraRotDetailer",
    "CameraSmallRotPerObservationListDetailer",
    "ComCamGridDitherDetailer",
)

import numpy as np

from rubin_scheduler.scheduler.detailers import BaseDetailer
from rubin_scheduler.scheduler.utils import wrap_ra_dec
from rubin_scheduler.utils import (
    _approx_altaz2pa,
    _approx_ra_dec2_alt_az,
    bearing,
    ddf_locations,
    dest_latlon,
    gnomonic_project_tosky,
    rotation_converter,
)


class DitherDetailer(BaseDetailer):
    """
    make a uniform dither pattern. Offset by a maximum radius in a random
    direction. Mostly intended for DDF pointings, the BaseMarkovDF_survey
    class includes dithering for large areas.

    Parameters
    ----------
    max_dither : `float` (0.7)
        The maximum dither size to use (degrees).
    per_night : `bool` (True)
        If true, us the same dither offset for an entire night
    nnights : `int` (7305)
        The number of nights to pre-generate random dithers for


    """

    def __init__(self, max_dither=0.7, seed=42, per_night=True, nnights=7305):
        self.survey_features = {}

        self.current_night = -1
        self.max_dither = np.radians(max_dither)
        self.per_night = per_night
        self.rng = np.random.default_rng(seed)
        self.angles = self.rng.random(nnights) * 2 * np.pi
        self.radii = self.max_dither * np.sqrt(self.rng.random(nnights))
        self.offsets = (self.rng.random((nnights, 2)) - 0.5) * 2.0 * self.max_dither
        self.offset = None

    def _generate_offsets(self, n_offsets, night):
        if self.per_night:
            if night != self.current_night:
                self.current_night = night
                self.offset = self.offsets[night, :]
                angle = self.angles[night]
                radius = self.radii[night]
                self.offset = np.array([radius * np.cos(angle), radius * np.sin(angle)])
            offsets = np.tile(self.offset, (n_offsets, 1))
        else:
            angle = self.rng.random(n_offsets) * 2 * np.pi
            radius = self.max_dither * np.sqrt(self.rng.random(n_offsets))
            offsets = np.array([radius * np.cos(angle), radius * np.sin(angle)]).T

        return offsets

    def __call__(self, observation_list, conditions):
        if len(observation_list) == 0:
            return observation_list
        # Generate offsets in RA and Dec
        offsets = self._generate_offsets(len(observation_list), conditions.night)

        obs_array = np.concatenate(observation_list)
        new_ra, new_dec = gnomonic_project_tosky(
            offsets[:, 0], offsets[:, 1], obs_array["RA"], obs_array["dec"]
        )
        new_ra, new_dec = wrap_ra_dec(new_ra, new_dec)
        for i, obs in enumerate(observation_list):
            observation_list[i]["RA"] = new_ra[i]
            observation_list[i]["dec"] = new_dec[i]
        return observation_list


class EuclidDitherDetailer(BaseDetailer):
    """Directional dithering for Euclid DDFs

    Parameters
    ----------
    dither_bearing_dir : `list`
        A list with the dither amplitude in along the axis
        connecting the two positions. Default [-0.25, 1] in degrees.
    dither_bearing_perp : `list`
        A list with the dither amplitude perpendicular to the
        axis connecting the two positions. Default [-0.25, 1]
        in degrees.
    seed : `float`
        Random number seed to use (42).
    per_night : `bool`
        If dither shifts should be per night (default True),
        or if dithers should be every pointing (False).
    ra_a, dec_a, ra_b, dec_b : `float`
        Positions for the two field centers. Default None
        will load the positions from
        rubin_scheduler.utils.ddf_locations
    nnights : `int`
        Number of nights to generate dither positions for.
        Default 7305 (20 years).
    """

    def __init__(
        self,
        dither_bearing_dir=[-0.25, 1],
        dither_bearing_perp=[-0.25, 0.25],
        seed=42,
        per_night=True,
        ra_a=None,
        dec_a=None,
        ra_b=None,
        dec_b=None,
        nnights=7305,
    ):
        self.survey_features = {}

        default_locations = ddf_locations()

        if ra_a is None:
            self.ra_a = np.radians(default_locations["EDFS_a"][0])
        else:
            self.ra_a = np.radians(ra_a)
        if ra_b is None:
            self.ra_b = np.radians(default_locations["EDFS_b"][0])
        else:
            self.ra_b = np.radians(ra_b)
        if dec_a is None:
            self.dec_a = np.radians(default_locations["EDFS_a"][1])
        else:
            self.dec_a = np.radians(dec_a)
        if dec_b is None:
            self.dec_b = np.radians(default_locations["EDFS_b"][1])
        else:
            self.dec_b = np.radians(dec_b)

        self.dither_bearing_dir = np.radians(dither_bearing_dir)
        self.dither_bearing_perp = np.radians(dither_bearing_perp)

        self.bearing_atob = bearing(self.ra_a, self.dec_a, self.ra_b, self.dec_b)
        self.bearing_btoa = bearing(self.ra_b, self.dec_b, self.ra_a, self.dec_a)

        self.current_night = -1

        self.per_night = per_night
        self.shifted_ra_a = None
        self.shifted_dec_a = None
        self.shifted_ra_b = None
        self.shifted_dec_b = None

        rng = np.random.default_rng(seed)
        self.bearings_mag_1 = rng.uniform(
            low=self.dither_bearing_dir.min(),
            high=self.dither_bearing_dir.max(),
            size=nnights,
        )
        self.perp_mag_1 = rng.uniform(
            low=self.dither_bearing_perp.min(),
            high=self.dither_bearing_perp.max(),
            size=nnights,
        )

        self.bearings_mag_2 = rng.uniform(
            low=self.dither_bearing_dir.min(),
            high=self.dither_bearing_dir.max(),
            size=nnights,
        )
        self.perp_mag_2 = rng.uniform(
            low=self.dither_bearing_perp.min(),
            high=self.dither_bearing_perp.max(),
            size=nnights,
        )

    def _generate_offsets(self, n_offsets, night):
        if self.per_night:
            if night != self.current_night:
                self.current_night = night
                bearing_mag = self.bearings_mag_1[night]
                perp_mag = self.perp_mag_1[night]
                # Move point a along the bearings
                self.shifted_dec_a, self.shifted_ra_a = dest_latlon(
                    bearing_mag, self.bearing_atob, self.dec_a, self.ra_a
                )
                self.shifted_dec_a, self.shifted_ra_a = dest_latlon(
                    perp_mag,
                    self.bearing_atob + np.pi / 2.0,
                    self.shifted_dec_a,
                    self.shifted_ra_a,
                )

                # Shift the second position
                bearing_mag = self.bearings_mag_2[night]
                perp_mag = self.perp_mag_2[night]

                self.shifted_dec_b, self.shifted_ra_b = dest_latlon(
                    bearing_mag, self.bearing_btoa, self.dec_b, self.ra_b
                )
                self.shifted_dec_b, self.shifted_ra_b = dest_latlon(
                    perp_mag,
                    self.bearing_btoa + np.pi / 2.0,
                    self.shifted_dec_b,
                    self.shifted_ra_b,
                )
        else:
            ValueError("not implamented")

        return (
            self.shifted_ra_a,
            self.shifted_dec_a,
            self.shifted_ra_b,
            self.shifted_dec_b,
        )

    def __call__(self, observation_list, conditions):
        # Generate offsets in RA and Dec
        ra_a, dec_a, ra_b, dec_b = self._generate_offsets(len(observation_list), conditions.night)

        for i, obs in enumerate(observation_list):
            if obs[0]["scheduler_note"][-1] == "a":
                observation_list[i]["RA"] = ra_a
                observation_list[i]["dec"] = dec_a
            elif obs[0]["scheduler_note"][-1] == "b":
                observation_list[i]["RA"] = ra_b
                observation_list[i]["dec"] = dec_b
            else:
                ValueError("observation note does not end in a or b.")
        return observation_list


class CameraRotDetailer(BaseDetailer):
    """
    Randomly set the camera rotation, either for each exposure, or per night.

    Parameters
    ----------
    max_rot : `float` (90.)
        The maximum amount to offset the camera (degrees)
    min_rot : `float` (90)
        The minimum to offset the camera (degrees)
    per_night : `bool` (True)
        If True, only set a new offset per night. If False, randomly
        rotates every observation.
    telescope : `str`
        Telescope name. Options of "rubin" or "auxtel". Default "rubin".
    """

    def __init__(self, max_rot=90.0, min_rot=-90.0, per_night=True, seed=42, nnights=7305, telescope="rubin"):
        self.survey_features = {}

        self.current_night = -1
        self.max_rot = np.radians(max_rot)
        self.min_rot = np.radians(min_rot)
        self.range = self.max_rot - self.min_rot
        self.per_night = per_night
        self.rng = np.random.default_rng(seed)
        self.offsets = self.rng.random(nnights)

        self.offset = None
        self.rc = rotation_converter(telescope=telescope)

    def _generate_offsets(self, n_offsets, night):
        if self.per_night:
            if night != self.current_night:
                self.current_night = night
                self.offset = self.offsets[night] * self.range + self.min_rot
            offsets = np.ones(n_offsets) * self.offset
        else:
            self.rng = np.random.default_rng()
            offsets = self.rng.random(n_offsets) * self.range + self.min_rot

        return offsets

    def __call__(self, observation_list, conditions):
        # Generate offsets in camamera rotator
        offsets = self._generate_offsets(len(observation_list), conditions.night)

        for i, obs in enumerate(observation_list):
            alt, az = _approx_ra_dec2_alt_az(
                obs["RA"],
                obs["dec"],
                conditions.site.latitude_rad,
                conditions.site.longitude_rad,
                conditions.mjd,
            )
            obs_pa = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)
            obs["rotSkyPos"] = self.rc._rottelpos2rotskypos(offsets[i], obs_pa)
            obs["rotTelPos"] = offsets[i]

        return observation_list


class CameraSmallRotPerObservationListDetailer(BaseDetailer):
    """
    Randomly set the camera rotation for each observation list.

    Generates a small sequential offset for sequential visits
    in the same filter; adds a random offset for each filter change.

    Parameters
    ----------
    max_rot : `float`, optional
        The maximum amount to offset the camera (degrees).
        Default of 85 allows some padding for camera rotator.
    min_rot : `float`, optional
        The minimum to offset the camera (degrees)
        Default of -85 allows some padding for camera rotator.
    seed : `int`, optional
        Seed for random number generation (per night).
    per_visit_rot : `float`, optional
        Sequential rotation to add per visit.
    telescope : `str`, optional
        Telescope name. Options of "rubin" or "auxtel". Default "rubin".
        This is used to determine conversions between rotSkyPos and rotTelPos.
    """

    def __init__(self, max_rot=85.0, min_rot=-85.0, seed=42, per_visit_rot=0.0, telescope="rubin"):
        self.survey_features = {}

        self.current_night = -1
        self.max_rot = np.radians(max_rot)
        self.min_rot = np.radians(min_rot)
        self.rot_range = self.max_rot - self.min_rot
        self.seed = seed
        self.per_visit_rot = np.radians(per_visit_rot)
        self.offset = None
        self.rc = rotation_converter(telescope=telescope)

    def _generate_offsets_filter_change(self, filter_list, mjd, initial_offset):
        """Generate a random camera rotation for each filter change
        or add a small offset for each sequential observation.
        """
        mjd_hash = round(100 * (np.asarray(mjd).item() % 100))
        rng = np.random.default_rng(mjd_hash * self.seed)

        offsets = np.zeros(len(filter_list))

        # Find the locations of the filter changes
        filter_changes = np.where(np.array(filter_list[:-1]) != np.array(filter_list[1:]))[0]
        filter_changes = np.concatenate([np.array([-1]), filter_changes])
        # But add one because of counting and offsets above.
        filter_changes += 1
        # Count visits per filter in the sequence.
        nvis_per_filter = np.concatenate(
            [np.diff(filter_changes), np.array([len(filter_list) - 1 - filter_changes[-1]])]
        )
        # Set up the random rotator offsets for each filter change
        # This includes first rotation .. maybe not needed?
        for fchange_idx, nvis_f in zip(filter_changes, nvis_per_filter):
            rot_range = self.rot_range - self.per_visit_rot * nvis_f
            # At the filter change spot, update to random offset
            offsets[fchange_idx:] = rng.random() * rot_range + self.min_rot
            # After the filter change point, add incremental rotation
            # (we'll wipe this when we get to next fchange_idx)
            offsets[fchange_idx:] += self.per_visit_rot * np.arange(len(filter_list) - fchange_idx)

        offsets = np.where(offsets > self.max_rot, self.max_rot, offsets)
        return offsets

    def __call__(self, observation_list, conditions):
        # Generate offsets in camera rotator
        filter_list = [np.asarray(obs["filter"]).item() for obs in observation_list]
        offsets = self._generate_offsets_filter_change(filter_list, conditions.mjd, conditions.rot_tel_pos)

        for i, obs in enumerate(observation_list):
            alt, az = _approx_ra_dec2_alt_az(
                obs["RA"],
                obs["dec"],
                conditions.site.latitude_rad,
                conditions.site.longitude_rad,
                conditions.mjd,
            )
            obs_pa = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)
            obs["rotSkyPos"] = self.rc._rottelpos2rotskypos(offsets[i], obs_pa)
            obs["rotTelPos"] = offsets[i]

        return observation_list


class ComCamGridDitherDetailer(BaseDetailer):
    """
    Generate an offset pattern to synthesize a 2x2 grid of ComCam pointings.

    Parameters
    ----------
    rotTelPosDesired : `float`, (0.)
        The physical rotation angle of the camera rotator (degrees)
    scale : `float` (0.355)
        Half of the offset between grid pointing centers. (degrees)
    dither : `float` (0.05)
        Dither offsets within grid to fill chip gaps. (degrees)
    telescope : `str`, ("comcam")
        Telescope name. Default "comcam".
        This is used to determine conversions between rotSkyPos and rotTelPos.
    """

    def __init__(self, rotTelPosDesired=0.0, scale=0.355, dither=0.05, telescope="comcam"):
        self.rotTelPosDesired = np.radians(rotTelPosDesired)
        self.scale = np.radians(scale)
        self.dither = np.radians(dither)
        self.rc = rotation_converter(telescope=telescope)

    def _rotate(self, x, y, angle):
        x_rot = x * np.cos(angle) - y * np.sin(angle)
        y_rot = x * np.sin(angle) + y * np.cos(angle)
        return x_rot, y_rot

    def _generate_offsets(self, n_offsets, filter_list, rotSkyPos):
        # 2 x 2 pointing grid
        x_grid = np.array([-1.0 * self.scale, -1.0 * self.scale, self.scale, self.scale])
        y_grid = np.array([-1.0 * self.scale, self.scale, self.scale, -1.0 * self.scale])
        x_grid_rot, y_grid_rot = self._rotate(x_grid, y_grid, -1.0 * rotSkyPos)
        offsets_grid_rot = np.array([x_grid_rot, y_grid_rot]).T

        # Dither pattern within grid to fill chip gaps
        # Psuedo-random offsets
        x_dither = np.array(
            [
                0.0,
                -0.5 * self.dither,
                -1.25 * self.dither,
                1.5 * self.dither,
                0.75 * self.dither,
            ]
        )
        y_dither = np.array(
            [
                0.0,
                -0.75 * self.dither,
                1.5 * self.dither,
                1.25 * self.dither,
                -0.5 * self.dither,
            ]
        )
        x_dither_rot, y_dither_rot = self._rotate(x_dither, y_dither, -1.0 * rotSkyPos)
        offsets_dither_rot = np.array([x_dither_rot, y_dither_rot]).T

        # Find the indices of the filter changes
        filter_changes = np.where(np.array(filter_list[:-1]) != np.array(filter_list[1:]))[0]
        filter_changes = np.concatenate([np.array([-1]), filter_changes])
        filter_changes += 1

        offsets = []
        index_filter = 0
        for ii in range(0, n_offsets):
            if ii in filter_changes:
                # Reset the count after each filter change
                index_filter = 0

            index_grid = index_filter % 4
            index_dither = np.floor(index_filter / 4).astype(int) % 5
            offsets.append(offsets_grid_rot[index_grid] + offsets_dither_rot[index_dither])
            index_filter += 1

        return np.vstack(offsets)

    def __call__(self, observation_list, conditions):
        if len(observation_list) == 0:
            return observation_list

        filter_list = [np.asarray(obs["filter"]).item() for obs in observation_list]

        # Initial estimate of rotSkyPos corresponding to desired rotTelPos
        alt, az, pa = _approx_ra_dec2_alt_az(
            observation_list[0]["RA"],
            observation_list[0]["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
            return_pa=True,
        )
        rotSkyPos = self.rc._rottelpos2rotskypos(self.rotTelPosDesired, pa)

        # Generate offsets in RA and Dec
        offsets = self._generate_offsets(len(observation_list), filter_list, rotSkyPos)

        # Project offsets onto sky
        obs_array = np.concatenate(observation_list)
        new_ra, new_dec = gnomonic_project_tosky(
            offsets[:, 0], offsets[:, 1], obs_array["RA"], obs_array["dec"]
        )
        new_ra, new_dec = wrap_ra_dec(new_ra, new_dec)

        # Update observations
        for ii in range(0, len(observation_list)):
            observation_list[ii]["RA"] = new_ra[ii]
            observation_list[ii]["dec"] = new_dec[ii]

            alt, az, pa = _approx_ra_dec2_alt_az(
                new_ra[ii],
                new_dec[ii],
                conditions.site.latitude_rad,
                conditions.site.longitude_rad,
                conditions.mjd,
                return_pa=True,
            )
            observation_list[ii]["rotSkyPos"] = rotSkyPos
            observation_list[ii]["rotTelPos"] = self.rc._rotskypos2rottelpos(rotSkyPos, pa)

        return observation_list



import unittest
from inspect import signature

import numpy as np

from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.model_observatory import ModelObservatory


class TestConditions(unittest.TestCase):

    def test_conditions(self):
        # Generate a Conditions object with things filled out
        mo = ModelObservatory()
        mo_conditions = mo.return_conditions()

        # Fresh empty conditions
        conditions = Conditions()

        # Find the arguments that need to be set
        auxtel_args = signature(conditions.set_auxtel_info)

        to_pass = []
        for key in auxtel_args.parameters.keys():
            if key != "kwargs":
                to_pass.append(getattr(mo_conditions, key))

        conditions.set_auxtel_info(*to_pass)

        for key in auxtel_args.parameters.keys():
            if key != "kwargs":
                if np.size(getattr(conditions, key)) == 1:
                    assert getattr(conditions, key) == getattr(mo_conditions, key)
                else:
                    assert np.array_equal(
                        getattr(conditions, key), getattr(mo_conditions, key), equal_nan=True
                    )

        # Again for the maintel attrs
        conditions = Conditions()
        maintel_args = signature(conditions.set_maintel_info)

        to_pass = []
        for key in maintel_args.parameters.keys():
            if key != "kwargs":
                to_pass.append(getattr(mo_conditions, key))

        conditions.set_maintel_info(*to_pass)

        for key in maintel_args.parameters.keys():
            if key != "kwargs":
                if (np.size(getattr(conditions, key)) == 1) | (isinstance(getattr(conditions, key), list)):
                    assert getattr(conditions, key) == getattr(mo_conditions, key)
                else:
                    assert np.array_equal(
                        getattr(conditions, key), getattr(mo_conditions, key), equal_nan=True
                    )

        # check we can set some arbitrary attributes
        conditions = Conditions()
        conditions.set_attrs(mjd=62511, moon_alt=0.1)

        with self.assertWarns(Warning):
            conditions.set_attrs(not_an_attribute=10)


if __name__ == "__main__":
    unittest.main()



__all__ = ("ddf_locations_skycoord", "ddf_locations", "ddf_locations_pre3_5")

import astropy.units as u
from astropy.coordinates import SkyCoord


def ddf_locations_skycoord():
    """Return the DDF locations as a dict of SkyCoord values."""
    # The DDF locations here are from Neil Brandt's white paper
    # submitted in response to the 2018 Call for White Papers on observing
    # strategy
    # Document-30468 -- AGN-DDF-WP-02.pdf
    # The locations are chosen based on existing multi-wavelength
    # coverage, plus an offset to avoid the bright star Mira near XMM-LSS
    ddf = {}
    ddf["ELAISS1"] = SkyCoord("00:37:48 −44:01:30", unit=(u.hourangle, u.deg), frame="icrs")
    ddf["XMM_LSS"] = SkyCoord("02:22:18  −04:49:00", unit=(u.hourangle, u.deg), frame="icrs")
    ddf["ECDFS"] = SkyCoord("03:31:55  −28:07:00", unit=(u.hourangle, u.deg), frame="icrs")
    ddf["COSMOS"] = SkyCoord("10:00:26  +02:14:01", unit=(u.hourangle, u.deg), frame="icrs")
    ddf["EDFS_a"] = SkyCoord(ra=58.90 * u.deg, dec=-49.32 * u.deg, frame="icrs")
    ddf["EDFS_b"] = SkyCoord(ra=63.60 * u.deg, dec=-47.60 * u.deg, frame="icrs")
    return ddf


def ddf_locations():
    """Return the DDF locations as a dict in degrees."""
    result = ddf_locations_skycoord()
    for r in result:
        result[r] = (result[r].ra.deg, result[r].dec.deg)

    return result


def ddf_locations_pre3_5():
    """Return the DDF locations used for v1 to v3.4 simulations."""
    ddf = {}
    ddf["ELAISS1"] = (9.45, -44.0)
    ddf["XMM_LSS"] = (35.708333, -4 - 45 / 60.0)
    ddf["ECDFS"] = (53.125, -28.0 - 6 / 60.0)
    ddf["COSMOS"] = (150.1, 2.0 + 10.0 / 60.0 + 55 / 3600.0)
    ddf["EDFS_a"] = (58.90, -49.315)
    ddf["EDFS_b"] = (63.6, -47.60)
    return ddf



import os
import unittest

from astropy.time import Time, TimeDelta

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.site_models import ScheduledDowntimeData


class ScheduledDowntimeDataTest(unittest.TestCase):
    def setUp(self):
        self.th = Time("2020-01-01", format="isot", scale="tai")
        self.startofnight = -0.34
        self.downtime_db = os.path.join(get_data_dir(), "site_models", "scheduled_downtime.db")

    def test_basic_information_after_creation(self):
        downtime_data = ScheduledDowntimeData(self.th, start_of_night_offset=self.startofnight)
        self.assertEqual(downtime_data.scheduled_downtime_db, self.downtime_db)
        self.assertEqual(self.th + TimeDelta(self.startofnight, format="jd"), downtime_data.night0)
        downtime_data = ScheduledDowntimeData(self.th, start_of_night_offset=0)
        self.assertEqual(downtime_data.night0, self.th)

    def test_information_after_initialization(self):
        downtime_data = ScheduledDowntimeData(self.th, start_of_night_offset=self.startofnight)
        downtime_data.read_data()
        self.assertEqual(len(downtime_data.downtime), 31)
        # Check some of the downtime values.
        dnight = downtime_data.downtime["end"] - downtime_data.downtime["start"]
        self.assertEqual(dnight[0].jd, 7)
        self.assertEqual(downtime_data.downtime["activity"][0], "general maintenance")
        self.assertEqual(dnight[4].jd, 14)
        self.assertEqual(downtime_data.downtime["activity"][4], "recoat mirror")

    def test_call(self):
        downtime_data = ScheduledDowntimeData(self.th, start_of_night_offset=self.startofnight)
        downtime_data.read_data()
        downtimes = downtime_data()
        self.assertEqual(downtimes["activity"][4], "recoat mirror")


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

import rubin_scheduler.site_models as site_models
from rubin_scheduler.scheduler.model_observatory import ModelObservatory


class TestConstantWeather(unittest.TestCase):
    def test_all_constant_weather(self):
        test_seeing = 1.234
        test_clouds = 0.11
        test_wind_speed = 10.0
        test_wind_direction = np.radians(30.0)

        seeing_data = site_models.ConstantSeeingData(test_seeing)
        cloud_data = site_models.ConstantCloudData(test_clouds)
        wind_data = site_models.ConstantWindData(
            wind_speed=test_wind_speed,
            wind_direction=test_wind_direction,
        )

        model_observatory = ModelObservatory(
            seeing_data=seeing_data, cloud_data=cloud_data, wind_data=wind_data, no_sky=True
        )
        conditions = model_observatory.return_conditions()
        self.assertEqual(conditions.wind_direction, test_wind_direction)
        self.assertEqual(conditions.wind_speed, test_wind_speed)
        self.assertEqual(conditions.bulk_cloud, test_clouds)
        assert np.nanmin(conditions.fwhm_eff["g"]) > test_seeing


if __name__ == "__main__":
    unittest.main()



from .almanac import *
from .cloud_data import *
from .cloud_model import *
from .downtime_model import *
from .read_fields import *
from .scheduled_downtime_data import *
from .seeing_data import *
from .seeing_model import *
from .unscheduled_downtime_data import *
from .wind_data import *



import importlib
import lzma
import pickle
import unittest
import urllib
from pathlib import Path
from tempfile import TemporaryDirectory

from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.example import example_scheduler
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.utils import SURVEY_START_MJD

HAVE_LSST_RESOURCES = importlib.util.find_spec("lsst") and importlib.util.find_spec("lsst.resources")
if HAVE_LSST_RESOURCES:
    from lsst.resources import ResourcePath

    from rubin_scheduler.sim_archive.sim_archive import (
        check_opsim_archive_resource,
        compile_sim_metadata,
        make_sim_archive_cli,
        make_sim_archive_dir,
        read_archived_sim_metadata,
        read_sim_metadata_from_hdf,
        transfer_archive_dir,
        verify_compiled_sim_metadata,
    )


class TestSimArchive(unittest.TestCase):
    @unittest.skipIf(not HAVE_LSST_RESOURCES, "No lsst.resources")
    def test_sim_archive(self):
        # Begin by running a short simulation
        sim_start_mjd = SURVEY_START_MJD
        sim_duration = 1  # days
        scheduler = example_scheduler(mjd_start=sim_start_mjd)
        scheduler.keep_rewards = True
        observatory = ModelObservatory(mjd_start=sim_start_mjd)

        # Record the state of the scheduler at the start of the sim.
        data_dir = TemporaryDirectory()
        data_path = Path(data_dir.name)

        scheduler_fname = data_path.joinpath("scheduler.pickle.xz")
        with lzma.open(scheduler_fname, "wb", format=lzma.FORMAT_XZ) as pio:
            pickle.dump(scheduler, pio)

        files_to_archive = {"scheduler": scheduler_fname}

        # Run the simulation
        sim_runner_kwargs = {
            "sim_start_mjd": sim_start_mjd,
            "sim_duration": sim_duration,
            "record_rewards": True,
        }

        observatory, scheduler, observations, reward_df, obs_rewards = sim_runner(
            observatory, scheduler, **sim_runner_kwargs
        )

        # Make the scratch sim archive
        make_sim_archive_dir(
            observations,
            reward_df=reward_df,
            obs_rewards=obs_rewards,
            in_files=files_to_archive,
            sim_runner_kwargs=sim_runner_kwargs,
            tags=["test"],
            label="test",
            data_path=data_path,
        )

        # Move the scratch sim archive to a test resource
        test_resource_dir = TemporaryDirectory()
        test_resource_uri = "file://" + test_resource_dir.name
        sim_archive_uri = transfer_archive_dir(data_dir.name, test_resource_uri)

        # Check the saved archive
        archive_check = check_opsim_archive_resource(sim_archive_uri)
        self.assertEqual(
            archive_check.keys(),
            set(
                [
                    "opsim.db",
                    "rewards.h5",
                    "scheduler.pickle.xz",
                    "obs_stats.txt",
                    "environment.txt",
                    "pypi.json",
                ]
            ),
        )
        for value in archive_check.values():
            self.assertTrue(value)

        # Read back the metadata
        archive_metadata = read_archived_sim_metadata(test_resource_uri)
        base = sim_archive_uri.dirname().geturl().removeprefix(test_resource_uri).rstrip("/").lstrip("/")
        expected_label = f"{base} test"
        self.assertEqual(archive_metadata[sim_archive_uri.geturl()]["label"], expected_label)

        # Cache the metadata
        test_compiled_metadata_uri = test_resource_uri + "/compiled_metadata_cache.h5"

        # Test reading from cached metadata
        compile_sim_metadata(test_resource_uri, test_compiled_metadata_uri)
        read_sim_metadata_from_hdf(test_compiled_metadata_uri)
        read_archived_sim_metadata(test_resource_uri, compilation_resource=test_compiled_metadata_uri)
        verify_compiled_sim_metadata(test_resource_uri, test_compiled_metadata_uri)

    @unittest.skipIf(not HAVE_LSST_RESOURCES, "No lsst.resources")
    @unittest.skipIf(importlib.util.find_spec("schedview") is None, "No schedview")
    def test_cli(self):
        test_resource_path = ResourcePath("resource://schedview/data/")
        with test_resource_path.join("sample_opsim.db").as_local() as local_rp:
            opsim = urllib.parse.urlparse(local_rp.geturl()).path

        with test_resource_path.join("sample_rewards.h5").as_local() as local_rp:
            rewards = urllib.parse.urlparse(local_rp.geturl()).path

        with test_resource_path.join("sample_scheduler.pickle.xz").as_local() as local_rp:
            scheduler = urllib.parse.urlparse(local_rp.geturl()).path

        with TemporaryDirectory() as test_archive_dir:
            test_archive_uri = f"file://{test_archive_dir}/"
            make_sim_archive_cli(
                "Test",
                opsim,
                "--rewards",
                rewards,
                "--scheduler",
                scheduler,
                "--archive_base_uri",
                test_archive_uri,
            )



__all__ = ("generate_ddf_scheduled_obs",)

import os
import warnings

import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler.utils import ScheduledObservationArray
from rubin_scheduler.site_models import Almanac
from rubin_scheduler.utils import SURVEY_START_MJD, calc_season, ddf_locations


def ddf_slopes(ddf_name, raw_obs, night_season, season_seq=30, min_season_length=0 / 365.25):
    """
    Let's make custom slopes for each DDF

    Parameters
    ----------
    ddf_name : `str`
       The DDF name to use
    raw_obs : `np.array`, (N,)
        An array with values of 1 or zero. One element per night, value of
        1 indicates the night is during an active observing season.
    night_season : `np.array`, (N,)
        An array of floats with the fractional season value.
        Season values range from 0-1 for the first season;
        0.5 would be the "peak" of seasonal visibility.
        These should ONLY be the night the field should be 'active'
        (take out the season ends first).
    season_seq : `int`, optional
        Number of sequences to try to place into each season.
        Default of 30.
    min_season_length : `float`, optional
        The minimum season length to allow before discarding the
        season. Default of 0 currently doesn't discard any seasons,
        on the basis that most of the short season concerns will be
        early in the survey and we'd rather have some observations early
        than wait.
    """

    # OK, so 258 sequences is ~1% of the survey
    # so a 25.8 sequences is a 0.1% season
    # COSMOS is going to be 0.7% for 3 years, then 0.175 for the rest.

    int_season = np.floor(night_season)

    n_season = night_season[np.where(raw_obs)]
    r_season = int_season[np.where(raw_obs)]
    season_list = np.unique(r_season)

    # Calculate season length for each season
    # In general this should be the same each season except for first and last
    season_length = np.zeros(len(season_list), float)
    for i, season in enumerate(season_list):
        match = np.where(r_season == season)
        season_length[i] = n_season[match].max() - n_season[match].min()

    # Determine goal number of sequences in each season.
    season_vals = np.ones(len(season_list), float) * season_seq
    # Adjust other seasons, relative to the max season length.
    season_vals = season_vals * season_length / np.max(season_length)
    # EXCEPT - throw out seasons which are too short
    too_short = np.where(season_length < min_season_length)
    season_vals[too_short] = 0

    # Add extra adjustment for COSMOS to boost visits in seasons 0-3
    # Also boost -1 season, if it's not too short
    boost_early_years = 5
    if ddf_name == "COSMOS":
        early_season = np.where(season_list == -1)[0]
        if len(early_season) > 0:
            max_n_seq_early = (season_length[early_season][0] * 365.25) / 3
            # Choose the minimum of - sequence every third night (?)
            # or the standard season sequence * boost_early_years.
            season_vals[early_season] = np.min([max_n_seq_early, season_seq * boost_early_years])
        first_full_two_seasons = np.where((season_list == 0) | (season_list == 1))
        season_vals[first_full_two_seasons] *= boost_early_years
        third_season = np.where(season_list == 2)
        season_vals[third_season] *= boost_early_years / 2.5

    if ddf_name == "EDFS_b":
        # EDFS_b ddf visits are allocated some other way
        season_vals = season_vals * 0

    # Round the season_vals -- we're looking for integer numbers of sequences
    season_vals = np.round(season_vals)

    # assign cumulative values
    cumulative_desired = np.zeros(raw_obs.size, dtype=float)
    for i, season in enumerate(season_list):
        this_season = np.where(int_season == season)
        # raw_obs gives a way to scale the season_vals across the nights
        # in the season -- raw_obs = 1 for peak of season, 0 when
        # beyond the season_unobs_frac, and a fraction for low_season_frac
        cumulative = np.cumsum(raw_obs[this_season])
        if cumulative.max() > 0:
            cumulative = cumulative / cumulative.max() * season_vals[i]
            cumulative_desired[this_season] = cumulative + np.max(cumulative_desired)

    # print(ddf_name, season_vals, cumulative_desired.max())

    return cumulative_desired


def match_cumulative(cumulative_desired, mask=None, no_duplicate=True):
    """Generate a schedule that tries to match the desired cumulative
    distribution given a mask

    Parameters
    ----------
    cumulative_desired : `np.array`, float
        An array with the cumulative number of desired observations.
        Elements  are assumed to be evenly spaced.
    mask : `np.array`, bool or int (None)
        Set to zero for indices that cannot be scheduled
    no_duplicate : `bool` (True)
        If True, only 1 event can be scheduled per element

    Returns
    -------
    schedule : `np.array`
        The resulting schedule, with values marking number of events
        in that cell.
    """

    rounded_desired = np.round(cumulative_desired)
    sched = cumulative_desired * 0
    if mask is None:
        mask = np.ones(sched.size)

    valid = np.where(mask > 0)[0].tolist()
    x = np.arange(sched.size)

    drd = np.diff(rounded_desired)
    step_points = np.where(drd > 0)[0] + 1

    # would be nice to eliminate this loop, but it's not too bad.
    # can't just use searchsorted on the whole array, because then there
    # can be duplicate values, and array[[n,n]] = 1 means that extra
    # match gets lost.
    for indx in step_points:
        left = np.searchsorted(x[valid], indx)
        right = np.searchsorted(x[valid], indx, side="right")
        d1 = indx - left
        d2 = right - indx
        if d1 < d2:
            sched_at = left
        else:
            sched_at = right

        # If we are off the end
        if sched_at >= len(valid):
            sched_at -= 1

        sched[valid[sched_at]] += 1
        if no_duplicate:
            valid.pop(sched_at)

    return sched


def optimize_ddf_times(
    ddf_name,
    ddf_RA,
    ddf_grid,
    sun_limit=-18,
    sequence_time=60.0,
    airmass_limit=2.5,
    sky_limit=None,
    g_depth_limit=23.5,
    season_unobs_frac=0.2,
    low_season_frac=0,
    low_season_rate=0.3,
    mjd_start=SURVEY_START_MJD,
):
    """

    Parameters
    ----------
    ddf_name : `str`
        The name of the DDF, used to identify visits scheduled for each DDF.
    ddf_RA : `float`
        The RA of the DDF, used to calculate the season values. In degrees.
    ddf_grid : `np.array`
        An array with info for the DDFs. Generated by the
        rubin_scheduler.scheduler/surveys/generate_ddf_grid.py` script
        The time spacing in this is approximately 15 or 30 minutes,
        and includes visits which may be during twilight.
    sun_limit : `float`, optional
        The maximum sun altitude allowed when prescheduling DDF visits.
        In degrees.
    sequence_time : `float`, optional
        Expected time for each DDF sequence, used to avoid hitting the
        sun_limit (running DDF visits into twilight). In minutes.
    airmass_limit : `float`, optional
        The maximum airmass allowed when prescheduling DDF visits.
    sky_limit : `float`, optional
        The maximum skybrightness allowed when prescheduling DDF visits.
        This is a skybrightness limit in g band (mags).
        Default None imposes no limit.
    g_depth_limit : `float`, optional
        The minimum g band five sigma depth limit allowed when prescheduling
        DDF visits. This is a depth limit in g band (mags).
        The depth is calculated using skybrightness from skybrightness_pre,
        a nominal FWHM_500 seeing at zenith of 0.7" (resulting in airmass
        dependent seeing) and exposure time.
        Default 23.5. Set to None for no limit.
    season_unobs_frac : `float`, optional
        Defines the end of the range of the prescheduled observing season.
        season runs from 0 (sun's apparent position is at the RA of the DDF)
        to 1 (sun returns to an apparent position in the RA of the DDF).
        The scheduled season runs from:
        season_unobs_frac < season < (1-season_unobs_fract)
    low_season_frac : `float`, optional
        Defines the end of the range of the "low cadence" prescheduled
        observing season.
        The "standard cadence" season runs from:
        low_season_frac < season < (1 - low_season_frac)
        For an 'accordian' style DDF with fewer observations near
        the ends of the season, set this to a value larger than
        `season_unobs_frac`. Values smaller than `season_unobs_frac`
        will result in DDFs with a constant rate throughout the season.
    low_season_rate : `float`, optional
        Defines the rate to use within the low cadence portion
        of the season. During the standard season, the 'rate' is 1.
        This is used in `ddf_slopes` to define the desired number of
        cumulative observations for each DDF over time.
    mjd_start : `float`, optional
        The MJD of the start of the survey. Used to identify the
        starting point when counting seasons.
        Default SURVEY_START_MJD.
    """
    # Convert sun_limit and sequence_time to values expected internally.
    sun_limit = np.radians(sun_limit)
    sequence_time = sequence_time / 60.0 / 24.0  # to days

    # Calculate the night value for each grid point.
    almanac = Almanac(mjd_start=ddf_grid["mjd"].min())
    almanac_indx = almanac.mjd_indx(ddf_grid["mjd"])
    night = almanac.sunsets["night"][almanac_indx]

    ngrid = ddf_grid["mjd"].size

    # Set the sun mask.
    sun_mask = np.ones(ngrid, dtype=int)
    sun_mask[np.where(ddf_grid["sun_alt"] >= sun_limit)] = 0
    # expand sun mask backwards by the sequence time.
    n_back = np.ceil(sequence_time / (ddf_grid["mjd"][1] - ddf_grid["mjd"][0])).astype(int)
    shadow_indx = np.where(sun_mask == 0)[0] - n_back
    shadow_indx = shadow_indx[np.where(shadow_indx >= 0)]
    sun_mask[shadow_indx] = 0

    # Set the airmass mask.
    airmass_mask = np.ones(ngrid, dtype=int)
    airmass_mask[np.where(ddf_grid["%s_airmass" % ddf_name] >= airmass_limit)] = 0

    # Set the sky_limit mask, if provided.
    sky_mask = np.ones(ngrid, dtype=int)
    if sky_limit is not None:
        sky_mask[np.where(ddf_grid["%s_sky_g" % ddf_name] <= sky_limit)] = 0
        sky_mask[np.where(np.isnan(ddf_grid["%s_sky_g" % ddf_name]) == True)] = 0

    # Set the m5 / g_depth_limit mask if provided.
    m5_mask = np.zeros(ngrid, dtype=bool)
    m5_mask[np.isfinite(ddf_grid["%s_m5_g" % ddf_name])] = 1
    if g_depth_limit is not None:
        m5_mask[np.where(ddf_grid["%s_m5_g" % ddf_name] < g_depth_limit)] = 0

    # Combine the masks.
    big_mask = sun_mask * airmass_mask * sky_mask * m5_mask

    # Identify which nights are useful to preschedule DDF visits.
    potential_nights = np.unique(night[np.where(big_mask > 0)])
    # prevent a repeat sequence in a night
    unights, indx = np.unique(night, return_index=True)
    night_mjd = ddf_grid["mjd"][indx]

    # Calculate season values for each night.
    night_season = calc_season(ddf_RA, night_mjd, mjd_start)
    # Mod by 1 to turn the season value in each night a simple 0-1 value
    season_mod = night_season % 1
    # Remove the portions of the season beyond season_unobs_frac.
    # Note that the "season" does include times where the field wouldn't
    # really have been observable (depending on declination).
    # Small values of season_unobs_frac may not influence the usable season.
    out_season = np.where((season_mod < season_unobs_frac) | (season_mod > (1.0 - season_unobs_frac)))
    # Identify the low-season times.
    low_season = np.where((season_mod < low_season_frac) | (season_mod > (1.0 - low_season_frac)))

    # Turn these into the 'rate scale' per night that goes to `ddf_slopes`
    raw_obs = np.ones(unights.size)
    raw_obs[out_season] = 0
    raw_obs[low_season] = low_season_rate

    cumulative_desired = ddf_slopes(ddf_name, raw_obs, night_season)

    # Identify which nights (only scheduling 1 sequence per night)
    # would be usable, based on the masks above.
    night_mask = unights * 0
    night_mask[potential_nights] = 1
    # Calculate expected cumulative sequence schedule.
    unight_sched = match_cumulative(cumulative_desired, mask=night_mask)
    cumulative_sched = np.cumsum(unight_sched)

    nights_to_use = unights[np.where(unight_sched == 1)]

    # For each night, find the best time in the night and preschedule the DDF.
    # XXX--probably need to expand this part to resolve the times when
    # multiple things get scheduled
    mjds = []
    for night_check in nights_to_use:
        in_night = np.where((night == night_check) & (np.isfinite(ddf_grid["%s_m5_g" % ddf_name])))[0]
        m5s = ddf_grid["%s_m5_g" % ddf_name][in_night]
        # we could intorpolate this to get even better than 15 min
        # resolution on when to observe
        max_indx = np.where(m5s == m5s.max())[0].min()
        mjds.append(ddf_grid["mjd"][in_night[max_indx]])

    return mjds, night_mjd, cumulative_desired, cumulative_sched


def generate_ddf_scheduled_obs(
    data_file=None,
    flush_length=2,
    mjd_tol=15,
    expt=30.0,
    alt_min=25,
    alt_max=85,
    HA_min=21.0,
    HA_max=3.0,
    sun_alt_max=-18,
    moon_min_distance=25.0,
    dist_tol=3.0,
    nvis_master=[8, 10, 20, 20, 24, 18],
    filters="ugrizy",
    nsnaps=[1, 2, 2, 2, 2, 2],
    mjd_start=SURVEY_START_MJD,
    survey_length=10.0,
    sequence_time=60.0,
    season_unobs_frac=0.2,
    low_season_frac=0,
    low_season_rate=0.3,
):
    """

    Parameters
    ----------
    data_file : `path` (None)
        The data file to use for DDF airmass, m5, etc. Defaults to
        using whatever is in rubin_sim_data/scheduler directory.
    flush_length : `float` (2)
        How long to keep a scheduled observation around before it is
        considered failed and flushed (days).
    mjd_tol : `float` (15)
        How close an observation must be in time to be considered
        matching a scheduled observation (minutes).
    expt : `float` (30)
        Total exposure time per visit (seconds).
    alt_min/max : `float` (25, 85)
        The minimum and maximum altitudes to permit observations to
        happen (degrees).
    HA_min/max : `float` (21, 3)
        The hour angle limits to permit observations to happen (hours).
    moon_min_distance : `float`
        The minimum distance to demand from the moon (degrees).
    dist_tol : `float` (3)
        The distance tolerance for a visit to be considered matching a
        scheduled observation (degrees).
    nvis_master : list of ints ([8, 10, 20, 20, 24, 18])
        The number of visits to make per filter
    filters : `str` (ugrizy)
        The filter names.
    nsnaps : `list of ints` ([1, 2, 2, 2, 2, 2])
        The number of snaps to use per filter
    mjd_start : `float`
        Starting MJD of the survey. Default None, which calls
        rubin_sim.utils.SURVEY_START_MJD
    survey_length : `float`
        Length of survey (years). Default 10.
    sequence_time : `float`, optional
        Expected time for each DDF sequence, used to avoid hitting the
        sun_limit (running DDF visits into twilight). In minutes.
    season_unobs_frac : `float`, optional
        Defines the end of the range of the prescheduled observing season.
        season runs from 0 (sun's apparent position is at the RA of the DDF)
        to 1 (sun returns to an apparent position in the RA of the DDF).
        The scheduled season runs from:
        season_unobs_frac < season < (1-season_unobs_fract)
    low_season_frac : `float`, optional
        Defines the end of the range of the "low cadence" prescheduled
        observing season.
        The "standard cadence" season runs from:
        low_season_frac < season < (1 - low_season_frac)
        For an 'accordian' style DDF with fewer observations near
        the ends of the season, set this to a value larger than
        `season_unobs_frac`. Values smaller than `season_unobs_frac`
        will result in DDFs with a constant rate throughout the season.
    low_season_rate : `float`, optional
        Defines the rate to use within the low cadence portion
        of the season. During the standard season, the 'rate' is 1.
        This is used in `ddf_slopes` to define the desired number of
        cumulative observations for each DDF over time.
    """
    if data_file is None:
        data_file = os.path.join(get_data_dir(), "scheduler", "ddf_grid.npz")

    flush_length = flush_length  # days
    mjd_tol = mjd_tol / 60 / 24.0  # minutes to days
    expt = expt
    alt_min = np.radians(alt_min)
    alt_max = np.radians(alt_max)
    dist_tol = np.radians(dist_tol)
    sun_alt_max = np.radians(sun_alt_max)
    moon_min_distance = np.radians(moon_min_distance)

    ddfs = ddf_locations()
    ddf_data = np.load(data_file)
    ddf_grid = ddf_data["ddf_grid"].copy()

    mjd_max = mjd_start + survey_length * 365.25

    # check if our pre-computed grid is over the time range we think
    # we are scheduling for
    if (ddf_grid["mjd"].min() > mjd_start) | (ddf_grid["mjd"].max() < mjd_max):
        warnings.warn("Pre-computed DDF properties don't match requested survey times")

    in_range = np.where((ddf_grid["mjd"] >= mjd_start) & (ddf_grid["mjd"] <= mjd_max))
    ddf_grid = ddf_grid[in_range]

    all_scheduled_obs = []
    for ddf_name in ["ELAISS1", "XMM_LSS", "ECDFS", "COSMOS", "EDFS_a"]:
        print("Optimizing %s" % ddf_name)

        mjds = optimize_ddf_times(
            ddf_name,
            ddfs[ddf_name][0],
            ddf_grid,
            season_unobs_frac=season_unobs_frac,
            sequence_time=sequence_time,
            low_season_frac=low_season_frac,
            low_season_rate=low_season_rate,
        )[0]
        for mjd in mjds:
            for filtername, nvis, nexp in zip(filters, nvis_master, nsnaps):
                if "EDFS" in ddf_name:
                    obs = ScheduledObservationArray(n=int(nvis / 2))
                    obs["RA"] = np.radians(ddfs[ddf_name][0])
                    obs["dec"] = np.radians(ddfs[ddf_name][1])
                    obs["mjd"] = mjd
                    obs["flush_by_mjd"] = mjd + flush_length
                    obs["exptime"] = expt
                    obs["filter"] = filtername
                    obs["nexp"] = nexp
                    obs["scheduler_note"] = "DD:%s" % ddf_name
                    obs["target_name"] = "DD:%s" % ddf_name

                    obs["mjd_tol"] = mjd_tol
                    obs["dist_tol"] = dist_tol
                    # Need to set something for HA limits
                    obs["HA_min"] = HA_min
                    obs["HA_max"] = HA_max
                    obs["alt_min"] = alt_min
                    obs["alt_max"] = alt_max
                    obs["sun_alt_max"] = sun_alt_max
                    all_scheduled_obs.append(obs)

                    obs = ScheduledObservationArray(n=int(nvis / 2))
                    obs["RA"] = np.radians(ddfs[ddf_name.replace("_a", "_b")][0])
                    obs["dec"] = np.radians(ddfs[ddf_name.replace("_a", "_b")][1])
                    obs["mjd"] = mjd
                    obs["flush_by_mjd"] = mjd + flush_length
                    obs["exptime"] = expt
                    obs["filter"] = filtername
                    obs["nexp"] = nexp
                    obs["scheduler_note"] = "DD:%s" % ddf_name.replace("_a", "_b")
                    obs["target_name"] = "DD:%s" % ddf_name.replace("_a", "_b")
                    obs["science_program"] = "DD"
                    obs["observation_reason"] = "FBS"

                    obs["mjd_tol"] = mjd_tol
                    obs["dist_tol"] = dist_tol
                    # Need to set something for HA limits
                    obs["HA_min"] = HA_min
                    obs["HA_max"] = HA_max
                    obs["alt_min"] = alt_min
                    obs["alt_max"] = alt_max
                    obs["sun_alt_max"] = sun_alt_max
                    obs["moon_min_distance"] = moon_min_distance
                    all_scheduled_obs.append(obs)

                else:
                    obs = ScheduledObservationArray(n=nvis)
                    obs["RA"] = np.radians(ddfs[ddf_name][0])
                    obs["dec"] = np.radians(ddfs[ddf_name][1])
                    obs["mjd"] = mjd
                    obs["flush_by_mjd"] = mjd + flush_length
                    obs["exptime"] = expt
                    obs["filter"] = filtername
                    obs["nexp"] = nexp
                    obs["scheduler_note"] = "DD:%s" % ddf_name
                    obs["target_name"] = "DD:%s" % ddf_name
                    obs["science_program"] = "DD"
                    obs["observation_reason"] = "FBS"

                    obs["mjd_tol"] = mjd_tol
                    obs["dist_tol"] = dist_tol
                    # Need to set something for HA limits
                    obs["HA_min"] = HA_min
                    obs["HA_max"] = HA_max
                    obs["alt_min"] = alt_min
                    obs["alt_max"] = alt_max
                    obs["sun_alt_max"] = sun_alt_max
                    obs["moon_min_distance"] = moon_min_distance
                    all_scheduled_obs.append(obs)

    result = np.concatenate(all_scheduled_obs)
    return result



__all__ = ["PointingsSurvey"]

import healpy as hp
import numpy as np
import pandas as pd

from rubin_scheduler.scheduler.detailers import ParallacticRotationDetailer
from rubin_scheduler.scheduler.utils import IntRounded
from rubin_scheduler.skybrightness_pre import dark_m5
from rubin_scheduler.utils import _angular_separation, _approx_ra_dec2_alt_az

from .base_survey import BaseSurvey


class PointingsSurvey(BaseSurvey):
    """Survey object for managing a set list of potential pointings
    without specified observing times. Does not follow the usual
    Survey class API by not using BasisFunction objects.

    Parameters
    ----------
    observations : `np.array`
        An array of observations, from e.g.,
        rubin_scheduler.scheduler.utils.ObservationArray
        expect "RA", "dec", and "note" to be filled, other columns ignored.
    gap_min : `float`
        The minimum gap to force between observations of the same
        spot (minutes)
    alt_min : `float`
        Altitude limit of the telescope (degrees). Default 20.
    alt_max : `float`
        Altitude limit of the telescope (degrees). Default 85.
    ha_max, ha_min : `float` (4,-4)
        hour angle limits (hours). Applied to all observations.
        Default 4,-4.
    weights : `dict`
        Dictionary with keys of method names and values of floats.
        Default of None uses {"visit_gap": 1.0, "balance_revisit": 1.0,
                              "m5diff": 1.0,
                              "wind_limit": 1.0, "slew_time": -1.0,
                              "ha_limit": 0, "alt_limit": 0,
                              "moon_limit": 0}
    wind_speed_maximum : `float`
        The maximum wind (m/s), mask any targets that would take more
        wind than that. Default 100 m/s.
    fiducial_FWHMEff : `float`
        A fiducial seeing value to use when computing dark sky depth.
        Default 0.7 arcsec.
    sun_alt_limit : `float`
        Have survey as infeasible when sun altitude is above the limit.
        Default -12 (degrees).
    track_notes_ngoal : `dict`
        If there are observations that should be tracked together
        (e.g., a sequence that should be observed together). Dict
        with keys of str and values of int.
    """

    def __init__(
        self,
        observations,
        gap_min=25.0,
        moon_dist_limit=30.0,
        weights=None,
        alt_max=85.0,
        alt_min=20.0,
        detailers=None,
        ha_max=4,
        ha_min=-4,
        wind_speed_maximum=100,
        fiducial_FWHMEff=0.7,
        sun_alt_limit=-12.0,
        track_notes_ngoal=None,
    ):
        # Not doing a super here, don't want to even have an nside defined.

        # Check that observations["scheduler_note"] are unique, otherwise
        # incoming observations will get double-counted
        if np.size(np.unique(observations["scheduler_note"])) != np.size(observations):
            raise ValueError("observations['scheduler_note'] values are not unique")

        self.observations = observations
        self.gap_min = gap_min / 60.0 / 24.0  # to days
        self.moon_dist_limit = np.radians(moon_dist_limit)
        self.alt_max = np.radians(alt_max)
        self.alt_min = np.radians(alt_min)
        self.zeros = self.observations["RA"] * 0.0
        self.wind_speed_maximum = wind_speed_maximum
        self.dark_map = None
        self.fiducial_FWHMEff = fiducial_FWHMEff
        self.sun_alt_limit = np.radians(sun_alt_limit)
        self.tracking_notes = {}
        self.track_notes_ngoal = track_notes_ngoal
        if track_notes_ngoal is not None:
            for key in track_notes_ngoal:
                self.tracking_notes[key] = 0
        # Make mapping so it's fast to compute reward later
        # Note that user must be careful to keep sequence names unique
        self.sequence_mapping = {}
        for key in self.tracking_notes:
            self.sequence_mapping[key] = []
        for i, obs in enumerate(observations):
            for key in self.sequence_mapping:
                if key in obs["scheduler_note"]:
                    self.sequence_mapping[key].append(i)

        # convert hour angle limits to radians and 0-2pi
        self.ha_max = (np.radians(ha_max * 360 / 24) + 2 * np.pi) % (2 * np.pi)
        self.ha_min = (np.radians(ha_min * 360 / 24) + 2 * np.pi) % (2 * np.pi)

        # Arrays to track progress
        self.n_obs = np.zeros(self.zeros.size, dtype=int)
        self.last_observed = np.zeros(self.zeros.size, dtype=float)

        self.last_computed_reward = -1.0

        if weights is None:
            self.weights = {
                "visit_gap": 1.0,
                "balance_revisit": 1.0,
                "m5diff": 1.0,
                "slew_time": -1.0,
                "wind_limit": 0.0,
                "ha_limit": 0,
                "alt_limit": 0,
                "moon_limit": 0,
            }
        else:
            self.weights = weights

        self.scheduled_obs = None
        # If there's no detailers, add one to set rotation to
        # Parallactic angle
        if detailers is None:
            self.detailers = [ParallacticRotationDetailer()]
        else:
            self.detailers = detailers

    def _check_feasibility(self, conditions):
        """Check if the survey is feasable in the current conditions"""
        # Simple feasability check.
        result = True

        # if the sun is too high
        if IntRounded(conditions.sun_alt) > IntRounded(self.sun_alt_limit):
            return False

        reward = self.calc_reward_function(conditions)
        if not np.isfinite(reward):
            result = False
        return result

    def calc_reward_function(self, conditions):
        """Compute reward function using methods set by `weights`
        dict on init."""
        if self.last_computed_reward != conditions.mjd:
            self.alt, self.az = _approx_ra_dec2_alt_az(
                self.observations["RA"],
                self.observations["dec"],
                conditions.site.latitude_rad,
                conditions.site.longitude_rad,
                conditions.mjd,
                lmst=conditions.lmst,
            )

            self.ha = np.radians(conditions.lmst * 360.0 / 24.0) - self.observations["RA"]
            self.ha[np.where(self.ha < 0)] += 2.0 * np.pi

            self.reward = np.zeros(self.observations.size, dtype=float)
            # Apply all the weights to the reward
            # In theory, could track where the reward is already
            # NaN and not bother doing extra computations.
            for key in self.weights:
                self.reward += self.weights[key] * getattr(self, key)(conditions)

            self.last_computed_reward = conditions.mjd
        return np.nanmax(self.reward)

    def generate_observations_rough(self, conditions):
        """Calculate reward function and highest reward observation.
        This is usually called by `generate_observations` which will
        take the result and then apply detailers to them.
        """
        max_reward = self.calc_reward_function(conditions)
        # take the first one in the array if there's a tie
        # Could change logic to return multiple pointings
        winner = np.min(np.where(self.reward == max_reward)[0])
        return [self.observations[winner].copy().reshape(1)]

    def add_observation(self, observation, indx=None):
        """Let survey know about a completed observation."""
        # Check for a nore match
        indx = np.where(observation["scheduler_note"] == self.observations["scheduler_note"])[0]
        # Tracking arrays
        self.n_obs[indx] += 1
        self.last_observed[indx] = observation["mjd"]

        # If we are tracking n observations of some type:
        for key in self.tracking_notes:
            if key in observation["scheduler_note"]:
                self.tracking_notes[key] += 1

    def add_observations_array(self, observations_array_in, observations_hpid_in):
        """Like `add_observation`, but for a large array of completed
        observations."""
        for unote in np.unique(observations_array_in["scheduler_note"]):
            matching = np.where(observations_array_in["scheduler_note"] == unote)[0]
            indx = np.where(self.observations["scheduler_note"] == unote)[0]
            self.n_obs[indx] += np.size(matching)
            self.last_observed[indx] = observations_array_in["mjd"][matching].max()

            for key in self.tracking_notes:
                if key in unote:
                    self.tracking_notes[key] += np.size(matching)

    def ha_limit(self, conditions):
        """Apply hour angle limits."""
        result = self.zeros.copy()
        # apply hour angle limits
        result[
            np.where(
                (IntRounded(self.ha) > IntRounded(self.ha_max))
                & (IntRounded(self.ha) < IntRounded(self.ha_min))
            )
        ] = np.nan
        return result

    def alt_limit(self, conditions):
        """Apply altitude limits."""
        result = self.zeros.copy()
        result[np.where(IntRounded(self.alt) > IntRounded(self.alt_max))] = np.nan
        result[np.where(IntRounded(self.alt) < IntRounded(self.alt_min))] = np.nan
        return result

    def moon_limit(self, conditions):
        """Apply moon distanve limit."""
        result = self.zeros.copy()
        dists = _angular_separation(
            self.observations["RA"], self.observations["dec"], conditions.moon_ra, conditions.moon_dec
        )
        result[np.where(IntRounded(dists) < IntRounded(self.moon_dist_limit))] = np.nan
        return result

    def wind_limit(self, conditions):
        """Apply the wind limit."""
        result = self.zeros.copy()
        if conditions.wind_speed is None or conditions.wind_direction is None:
            return result
        wind_pressure = conditions.wind_speed * np.cos(self.az - conditions.wind_direction)
        result -= wind_pressure**2.0
        mask = IntRounded(wind_pressure) > IntRounded(self.wind_speed_maximum)
        result[mask] = np.nan

        return result

    def visit_gap(self, conditions):
        """Enforce a minimum visit gap."""
        diff = conditions.mjd - self.last_observed
        too_soon = np.where(IntRounded(diff) < IntRounded(self.gap_min))[0]
        result = self.zeros.copy()
        # Using NaN makes it a hard limit
        # could have a weight and just subtract from the reward
        result[too_soon] = np.nan
        return result

    def balance_revisit(self, conditions):
        """Code to balance revisiting different targets."""
        sum_obs = np.sum(self.n_obs)
        result = np.floor(1.0 + self.n_obs / sum_obs)
        result[np.where(self.n_obs == 0)] = 1
        return result

    def slew_time(self, conditions):
        # Interpolate the conditions slewtime map to our positions
        # Could do something more sophisticated, but this is probably fine.
        result = hp.get_interp_val(
            conditions.slewtime,
            np.degrees(self.observations["RA"]),
            np.degrees(self.observations["dec"]),
            lonlat=True,
        )

        # Could do a filter check here and add a penalty for changing
        # filters

        return result

    def sequence_boost(self, conditions):
        """Boost the reward if a sequence is incomplete."""
        result = self.zeros.copy()
        for key in self.sequence_mapping:
            result[self.sequence_mapping[key]] = self.tracking_notes[key] % self.track_notes_ngoal[key]

        return result

    def _dark_map(self, conditions):
        """Generate the dark map if needed

        Constructs self.dark_map which is a dict with
        keys of filtername and values of HEALpix arrays
        that are the darkest expected 5-sigma limiting depth
        expected for that region of sky
        """
        self.dark_map = {}
        for filtername in np.unique(self.observations["filter"]):
            self.dark_map[filtername] = dark_m5(
                conditions.dec, filtername, conditions.site.latitude_rad, self.fiducial_FWHMEff
            )

    def m5diff(self, conditions):
        """Compute difference between current 5-sigma limiting depth and
        the depth at the same coordinates in "ideal" conditions.
        """
        if self.dark_map is None:
            self._dark_map(conditions)
        result = np.zeros(self.observations.size)
        for filtername in np.unique(self.observations["filter"]):
            indx = np.where(self.observations["filter"] == filtername)[0]
            diff_map = conditions.m5_depth[filtername] - self.dark_map[filtername]
            result[indx] = hp.get_interp_val(
                diff_map,
                np.degrees(self.observations["RA"][indx]),
                np.degrees(self.observations["dec"][indx]),
                lonlat=True,
            )

        return result

    def _reward_to_scalars(self, reward):
        scalar_reward = np.nanmax(reward)
        n_unmasked = np.sum(np.isfinite(reward))

        return scalar_reward, n_unmasked

    def make_reward_df(self, conditions, accum=True):
        """Create a pandas.DataFrame describing the reward from the
        survey."""

        feasibility = []
        max_rewards = []
        n_possibles = []
        accum_rewards = []
        accum_areas = []
        bf_label = []
        bf_class = []
        basis_weights = self.weights.values()

        short_labels = self.bf_short_labels

        tracking_reward = np.zeros(self.observations.size, dtype=float)

        for method_name in self.weights:
            bf_label.append(short_labels[method_name])
            bf_class.append(None)
            bf_reward = self.weights[method_name] * getattr(self, method_name)(conditions)
            max_reward, n_possible = self._reward_to_scalars(bf_reward)

            if np.isnan(max_reward) | (n_possible == 0):
                this_feasibility = False
            else:
                this_feasibility = True

            feasibility.append(this_feasibility)
            max_rewards.append(max_reward)
            n_possibles.append(n_possible)

            if accum:
                tracking_reward += bf_reward
                accum_reward, accum_area = self._reward_to_scalars(tracking_reward)
                accum_rewards.append(accum_reward)
                accum_areas.append(accum_area)

        reward_data = {
            "method": bf_label,
            "blank": bf_class,
            "feasible": feasibility,
            "max_reward": max_rewards,
            "n_possibles": n_possibles,
            "weight": basis_weights,
        }
        if accum:
            reward_data["max_accum_reward"] = accum_rewards
            reward_data["accum_n"] = accum_areas

        reward_df = pd.DataFrame(reward_data)

        return reward_df

    def reward_changes(self, conditions):
        """List the rewards for each basis function used by the survey."""

        reward_values = []

        for key in self.weights:
            reward_values.append(np.nanmax(getattr(self, key)(conditions)))

        names = list(self.weights.keys())

        return list(zip(names, reward_values))

    @property
    def bf_short_labels(self):
        long_labels = list(self.weights.keys())

        label_bases = [label.split(" @")[0] for label in long_labels]
        duplicated_labels = set([label for label in label_bases if label_bases.count(label) > 1])
        short_labels = []
        label_count = {k: 0 for k in duplicated_labels}
        for label_base in label_bases:
            if label_base in duplicated_labels:
                label_count[label_base] += 1
                short_labels.append(f"{label_base} {label_count[label_base]}")
            else:
                short_labels.append(label_base)

        label_map = dict(zip(long_labels, short_labels))

        return label_map



import unittest

import numpy as np
from astropy.time import Time

from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.example import (
    get_ideal_model_observatory,
    simple_field_survey,
    simple_greedy_survey,
    simple_pairs_survey,
    update_model_observatory_sunset,
)
from rubin_scheduler.scheduler.schedulers import ComCamFilterSched, CoreScheduler
from rubin_scheduler.utils import SURVEY_START_MJD


class TestSurveyConveniences(unittest.TestCase):

    def setUp(self) -> None:
        self.survey_start = np.floor(SURVEY_START_MJD) + 0.5
        self.day_obs_start = Time(self.survey_start, format="mjd", scale="utc").iso[:10]

    def test_model_observatory_conveniences(self):
        """Test the model observatory convenience functions."""

        # Just check that we can acquire a model observatory
        # and it is set up for the date expected.
        observatory = get_ideal_model_observatory(dayobs=self.day_obs_start, survey_start=self.survey_start)
        conditions = observatory.return_conditions()
        assert conditions.mjd == observatory.mjd
        # The model observatory automatically advanced to -12 deg sunset
        assert (conditions.mjd - self.survey_start) < 1

        newday = self.survey_start + 4
        new_dayobs = Time(newday, format="mjd", scale="utc").iso[:10]
        newday = Time(f"{new_dayobs}T12:00:00", format="isot", scale="utc").mjd
        observatory = get_ideal_model_observatory(dayobs=new_dayobs, survey_start=self.survey_start)
        conditions = observatory.return_conditions()
        assert (conditions.mjd - newday) < 1

        # And update observatory to sunset, using a filter scheduler
        # that only has 'g' available
        filter_sched = ComCamFilterSched(illum_bins=np.arange(0, 101, 100), loaded_filter_groups=(("g",)))
        observatory = update_model_observatory_sunset(observatory, filter_sched, twilight=-18)
        assert observatory.observatory.current_filter == "g"
        assert observatory.conditions.sun_alt < np.radians(18)

    def test_simple_greedy_survey(self):
        # Just test that it still instantiates and provides observations.
        observatory = get_ideal_model_observatory(dayobs=self.day_obs_start, survey_start=self.survey_start)
        greedy = [simple_greedy_survey(filtername="r")]
        scheduler = CoreScheduler(greedy)
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, filter_scheduler=None, sim_duration=0.7
        )
        # Current survey_start_mjd should produce ~1000 visits
        # but changing this to a short night could reduce the total number
        self.assertTrue(len(observations) > 650)
        self.assertTrue(observations["mjd"].max() - observations["mjd"].min() > 0.4)
        self.assertTrue(np.unique(observations["scheduler_note"]).size == 1)
        self.assertTrue(np.unique(observations["scheduler_note"])[0] == "simple greedy r")
        # Check that we tracked things appropriately
        self.assertTrue(
            len(observations) == scheduler.survey_lists[0][0].extra_features["ObsRecorded"].feature
        )
        self.assertTrue(
            np.abs(
                observations[-1]["mjd"]
                - scheduler.survey_lists[0][0].extra_features["LastObs"].feature["mjd"]
            )
            < 15 / 60 / 60 / 24
        )

    def test_simple_pairs_survey(self):
        # Just test that it still instantiates and provides observations.
        observatory = get_ideal_model_observatory(dayobs=self.day_obs_start, survey_start=self.survey_start)
        pairs = [simple_pairs_survey(filtername="r", filtername2="i")]
        scheduler = CoreScheduler(pairs)
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, filter_scheduler=None, sim_duration=0.7
        )
        # Current survey_start_mjd should produce over ~950 visits
        # but changing this to a short night could reduce the total number
        self.assertTrue(len(observations) > 650)
        self.assertTrue(observations["mjd"].max() - observations["mjd"].min() > 0.4)
        self.assertTrue(np.unique(observations["scheduler_note"]).size == 2)
        # Check that we tracked things appropriately
        self.assertTrue(
            len(observations) == scheduler.survey_lists[0][0].extra_features["ObsRecorded"].feature
        )
        self.assertTrue(
            np.abs(
                observations[-1]["mjd"]
                - scheduler.survey_lists[0][0].extra_features["LastObs"].feature["mjd"]
            )
            < 15 / 60 / 60 / 24
        )

    def test_simple_field_survey(self):
        # Just test that it still instantiates and provides observations.
        observatory = get_ideal_model_observatory(dayobs=self.day_obs_start, survey_start=self.survey_start)
        # This field ought to be observable at our current survey_start
        ra = 150
        dec = 2.2
        field_name = "almost_cosmos"
        field = [
            simple_field_survey(
                field_ra_deg=ra, field_dec_deg=dec, field_name=field_name, science_program="BLOCK-TEST"
            )
        ]
        # Add a greedy survey backup because of the visit gap requirement in
        # the default field survey
        greedy = [simple_greedy_survey(filtername="r")]
        scheduler = CoreScheduler([field, greedy])
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, filter_scheduler=None, sim_duration=0.7
        )
        # Current survey_start_mjd should produce over ~950 visits
        # but changing this to a short night could reduce the total number
        self.assertTrue(len(observations) > 650)
        self.assertTrue(observations["mjd"].max() - observations["mjd"].min() > 0.4)
        # Check some information about the observation notes and names
        self.assertTrue(np.unique(observations["scheduler_note"]).size == 2)
        self.assertTrue("almost_cosmos" in observations["scheduler_note"])
        self.assertTrue("almost_cosmos" in observations["target_name"])
        # Check that the field survey got lots of visits
        field_obs = observations[np.where(observations["target_name"] == "almost_cosmos")]
        self.assertTrue(field_obs.size > 200)
        self.assertTrue(np.all(field_obs["science_program"] == "BLOCK-TEST"))
        self.assertTrue(field[0].extra_features["ObsRecorded_note"].feature == field_obs.size)
        self.assertTrue(
            np.abs(field_obs[-1]["mjd"] - field[0].extra_features["LastObs_note"].feature["mjd"])
            < 15 / 60 / 60 / 24
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

import rubin_scheduler.utils as utils
from rubin_scheduler.scheduler.model_observatory import ModelObservatory, acc_time, jerk_time


class KindaClouds:
    """Dummy class that always sets the clouds level to a unique float"""

    def __call__(self, mjd):
        return 0.23659


class ArbSeeing:
    """Dummy class to always return a specific seeing value"""

    def __call__(self, mjd):
        fwhm_500 = 1.756978
        return fwhm_500


class TestModelObservatory(unittest.TestCase):
    def test_ideal(self):
        """test that we can set ideal conditions"""
        mjd_start = utils.SURVEY_START_MJD
        mo_default = ModelObservatory(mjd_start=mjd_start)

        mo_ideal = ModelObservatory(
            mjd_start=mjd_start, cloud_data="ideal", seeing_data="ideal", downtimes="ideal"
        )

        cond_default = mo_default.return_conditions()
        cond_ideal = mo_ideal.return_conditions()

        # Should be at ideal seeing
        assert mo_ideal.fwhm_500 == 0.7
        assert mo_default.fwhm_500 != 0.7

        assert cond_ideal.bulk_cloud == 0
        assert cond_default.bulk_cloud > 0

        mjd_down = mo_default.downtimes["start"][0] + 0.01

        assert ~mo_default.check_up(mjd_down)[0]
        assert mo_ideal.check_up(mjd_down)[0]

    def test_time_jumps(self):
        """test that we jump forward properly with downtime
        and daytime.
        """
        mjd_start = utils.SURVEY_START_MJD
        mo = ModelObservatory(mjd_start=mjd_start, cloud_data="ideal", seeing_data="ideal", downtimes="ideal")
        conditions = mo.return_conditions()

        # Set the time to right after sunset
        mo.mjd = conditions.sun_n18_setting

        # Set a short downtime of 20 minutes
        length = 20.0 / 60 / 24
        down_starts = [mo.mjd + length / 1000]
        down_ends = [mo.mjd + length]
        mo.downtimes = np.array(
            list(zip(down_starts, down_ends)),
            dtype=list(zip(["start", "end"], [float, float])),
        )

        in_up_time, new_time = mo.check_up(mo.mjd + length / 2)
        assert ~in_up_time
        assert new_time == down_ends[0]

        # Check that check_mjd advances the mjd to the correct time
        in_valid_time, new_time = mo.check_mjd(mo.mjd + length / 2)

        assert ~in_valid_time
        assert new_time == down_ends[0]

        # Check that we advance properly if in daytime.
        # 2 hours before sunset
        mo.mjd = conditions.sun_n12_setting - 2.0 / 24.0
        in_valid_time, new_time = mo.check_mjd(mo.mjd)
        assert ~in_valid_time
        assert new_time == conditions.sun_n12_setting

        # Check no time jump if in valid time
        good_time = down_ends[0] + 2.0 / 24.0
        mo.mjd = good_time
        in_valid_time, new_time = mo.check_mjd(mo.mjd)
        assert in_valid_time
        assert new_time == good_time

    def test_replace(self):
        """test that we can replace default downtimes, seeing, and clouds"""

        mjd_start = utils.SURVEY_START_MJD
        mo_default = ModelObservatory(mjd_start=mjd_start)
        # Never load too many nights of sky
        mo_default.sky_model.load_length = 10.0
        cond_default = mo_default.return_conditions()

        # Define new downtimes
        downtimes = np.zeros(2, dtype=list(zip(["start", "end"], [float, float])))
        downtimes["start"] = np.array([1, 10]) + mjd_start
        downtimes["end"] = np.array([2, 11]) + mjd_start

        seeing_data = ArbSeeing()
        cloud_data = KindaClouds()

        mo_new = ModelObservatory(
            mjd_start=mjd_start,
            seeing_data=seeing_data,
            cloud_data=cloud_data,
            downtimes=downtimes,
        )
        # Never load too many nights of sky
        mo_new.sky_model.load_length = 10.0
        cond_new = mo_new.return_conditions()

        # Make sure the internal downtimes are different
        assert ~np.array_equal(mo_default.downtimes, mo_new.downtimes)

        # Make sure seeing is not the same
        diff = cond_default.fwhm_eff["r"] - cond_new.fwhm_eff["r"]
        assert np.nanmin(np.abs(diff)) > 0

        # Make sure cloudyness level is not the same
        assert cond_default.bulk_cloud != cond_new.bulk_cloud

    def test_jerk(self):
        """Test that motion times using jerk are reasonable"""

        distance = 4.5  # degrees
        v_max = 1.75  # deg/s
        acc_max = 0.875  # deg/s/s
        jerk_max = 5.0  # deg/s/s/s

        # Test that adding jerk increases travel time
        t1 = jerk_time(distance, v_max, acc_max, jerk_max)
        t2 = acc_time(distance, v_max, acc_max)

        assert t1 > t2

        # Test that jerk of None reverts properly
        t1 = jerk_time(distance, v_max, acc_max, None)
        t2 = acc_time(distance, v_max, acc_max)

        assert t1 == t2

        # Test that large jerk is close to acceleration only time

        t1 = jerk_time(distance, v_max, acc_max, 1e9)
        t2 = acc_time(distance, v_max, acc_max)

        assert np.allclose(t1, t2, atol=1e-5)

        # Test that degrees or radians are the same
        t1 = jerk_time(distance, v_max, acc_max, jerk_max)
        t2 = jerk_time(np.radians(distance), np.radians(v_max), np.radians(acc_max), np.radians(jerk_max))

        assert np.allclose(t1, t2, atol=1e-7)

        # test if values are equal case.
        distance = 4.5  # degrees
        v_max = 3.5  # deg/s
        acc_max = 3.5  # deg/s/s
        jerk_max = 3.5  # deg/s/s/s

        t1 = jerk_time(distance, v_max, acc_max, jerk_max)


if __name__ == "__main__":
    unittest.main()



import numbers
import unittest

import numpy as np

from rubin_scheduler.utils import ZernikePolynomialGenerator, _FactorialGenerator


class FactorialTestCase(unittest.TestCase):
    def test_factorial(self):
        gen = _FactorialGenerator()
        ii = gen.evaluate(9)
        ans = 9 * 8 * 7 * 6 * 5 * 4 * 3 * 2
        self.assertEqual(ii, ans)

        ii = gen.evaluate(15)
        ans = 15 * 14 * 13 * 12 * 11 * 10 * 9 * 8 * 7 * 6 * 5 * 4 * 3 * 2
        self.assertEqual(ii, ans)

        ii = gen.evaluate(3)
        ans = 6
        self.assertEqual(ii, ans)

        self.assertEqual(gen.evaluate(0), 1)
        self.assertEqual(gen.evaluate(1), 1)


class ZernikeTestCase(unittest.TestCase):
    long_message = True

    @classmethod
    def setUpClass(cls):
        cls.d_r = 0.01
        cls.d_phi = 0.005 * np.pi
        r_grid = np.arange(0.0, 1.0, cls.d_r)
        phi_grid = np.arange(0.0, 2.0 * np.pi, cls.d_phi)
        mesh = np.meshgrid(r_grid, phi_grid)
        cls.r_grid = mesh[0].flatten()
        cls.phi_grid = mesh[1].flatten()

        r_grid = np.arange(0.0, 1.0, 0.1)
        phi_grid = np.arange(0.0, 2.0 * np.pi, 0.05 * np.pi)
        mesh = np.meshgrid(r_grid, phi_grid)
        cls.r_grid_small = mesh[0].flatten()
        cls.phi_grid_small = mesh[1].flatten()

    def test_orthogonality(self):
        """
        Test that ZernikePolynomialGenerator returns
        polynomials that are orthogonal on the unit disc
        """

        polynomials = {}
        z_gen = ZernikePolynomialGenerator()

        for n in range(3):
            for m in range(-n, n + 1, 2):
                vals = np.zeros(len(self.r_grid), dtype=float)
                for ii, (rr, pp) in enumerate(zip(self.r_grid, self.phi_grid)):
                    vals[ii] = z_gen.evaluate(rr, pp, n, m)
                nm_tuple = (n, m)
                polynomials[nm_tuple] = vals

        p_keys = list(polynomials.keys())
        for ii in range(len(p_keys)):
            p1_name = p_keys[ii]
            p1 = polynomials[p1_name]
            integral = (p1 * p1 * self.r_grid * self.d_r * self.d_phi).sum()
            normed_integral = integral / z_gen.norm(p1_name[0], p1_name[1])
            self.assertLess(np.abs(normed_integral - 1.0), 0.04)
            for jj in range(ii + 1, len(p_keys)):
                p2_name = p_keys[jj]
                p2 = polynomials[p2_name]
                dot = (p1 * p2 * self.r_grid * self.d_r * self.d_phi).sum()
                msg = "\n%s norm %e\n dot %e\n" % (p1_name, integral, dot)
                self.assertLess(np.abs(dot / integral), 0.01, msg=msg)

    def test_zeros(self):
        """
        Test that ZernikePolynomialGenerator returns zero
        when values of n and m require it.
        """
        rng = np.random.RandomState(88)
        z_gen = ZernikePolynomialGenerator()
        for n in range(4):
            for m in range(-(n - 1), n, 2):
                r = rng.random_sample()
                phi = rng.random_sample() * 2.0 * np.pi
                self.assertAlmostEqual(0.0, z_gen.evaluate(r, phi, n, m), 10)

    def test_r_greater_than_one(self):
        """
        Test that the expected error is raised if we try to evaluate
        the Zernike polynomial with r>1
        """
        z_gen = ZernikePolynomialGenerator()
        vv = z_gen.evaluate(1.2, 2.1, 2, 0)
        self.assertTrue(np.isnan(vv))
        vv = z_gen.evaluate(np.array([0.1, 0.5, 1.2]), np.array([0.1, 0.2, 0.3]), 2, -2)
        self.assertTrue(np.isnan(vv[2]))
        self.assertFalse(np.isnan(vv[0]))
        self.assertFalse(np.isnan(vv[1]))
        vv = z_gen.evaluate_xy(1.1, 1.2, 4, -2)
        self.assertTrue(np.isnan(vv))
        vv = z_gen.evaluate_xy(np.array([0.1, 0.2, 0.3]), np.array([0.1, 1.0, 0.1]), 4, 2)
        self.assertTrue(np.isnan(vv[1]))
        self.assertFalse(np.isnan(vv[0]))
        self.assertFalse(np.isnan(vv[2]))

    def test_array(self):
        """
        Test that ZernikePolynomialGenerator can handle arrays of inputs
        """
        z_gen = ZernikePolynomialGenerator()
        n = 2
        m = -2
        val_arr = z_gen.evaluate(self.r_grid_small, self.phi_grid_small, n, m)
        self.assertEqual(len(val_arr), len(self.r_grid_small))
        for ii, (rr, pp) in enumerate(zip(self.r_grid_small, self.phi_grid_small)):
            vv = z_gen.evaluate(rr, pp, n, m)
            self.assertAlmostEqual(vv, val_arr[ii], 14)

    def test_xy(self):
        """
        Test that ZernikePolynomialGenerator can handle Cartesian coordinates
        """
        n = 4
        m = 2
        z_gen = ZernikePolynomialGenerator()
        x = self.r_grid_small * np.cos(self.phi_grid_small)
        y = self.r_grid_small * np.sin(self.phi_grid_small)
        val_arr = z_gen.evaluate_xy(x, y, n, m)
        self.assertGreater(np.abs(val_arr).max(), 1.0e-6)
        for ii, (rr, pp) in enumerate(zip(self.r_grid_small, self.phi_grid_small)):
            vv = z_gen.evaluate(rr, pp, n, m)
            self.assertAlmostEqual(vv, val_arr[ii], 14)

    def test_xy_one_at_a_time(self):
        """
        Test that ZernikePolynomialGenerator can handle
        scalar Cartesian coordinates (as opposed to arrays
        of Cartesian coordinates)
        """
        n = 4
        m = 2
        z_gen = ZernikePolynomialGenerator()
        x = self.r_grid_small * np.cos(self.phi_grid_small)
        y = self.r_grid_small * np.sin(self.phi_grid_small)

        for ii in range(len(self.r_grid_small)):
            vv_r = z_gen.evaluate(self.r_grid_small[ii], self.phi_grid_small[ii], n, m)
            vv_xy = z_gen.evaluate_xy(x[ii], y[ii], n, m)
            self.assertAlmostEqual(vv_r, vv_xy, 14)
            self.assertIsInstance(vv_xy, numbers.Number)

    def test__zernike_origin(self):
        """
        Test that ZernikePolynomialGenerator is well-behaved
        at r=0
        """
        n = 4
        m = 2
        z_gen = ZernikePolynomialGenerator()
        ans = z_gen.evaluate(0.0, 1.2, n, m)
        self.assertEqual(ans, 0.0)
        ans = z_gen.evaluate(np.array([0.0, 0.0]), np.array([1.2, 2.1]), n, m)

        np.testing.assert_array_equal(ans, np.zeros(2, dtype=float))
        ans = z_gen.evaluate_xy(0.0, 0.0, n, m)
        self.assertEqual(ans, 0.0)
        ans = z_gen.evaluate_xy(np.zeros(2, dtype=float), np.zeros(2, dtype=float), n, m)
        np.testing.assert_array_equal(ans, np.zeros(2, dtype=float))

        n = 0
        m = 0
        ans = z_gen.evaluate(0.0, 1.2, n, m)
        self.assertEqual(ans, 1.0)
        ans = z_gen.evaluate(np.array([0.0, 0.0]), np.array([1.2, 2.1]), n, m)

        np.testing.assert_array_equal(ans, np.ones(2, dtype=float))
        ans = z_gen.evaluate_xy(0.0, 0.0, n, m)
        self.assertEqual(ans, 1.0)
        ans = z_gen.evaluate_xy(np.zeros(2, dtype=float), np.zeros(2, dtype=float), n, m)
        np.testing.assert_array_equal(ans, np.ones(2, dtype=float))


if __name__ == "__main__":
    unittest.main()



__all__ = ("GreedySurvey", "BlobSurvey")

import warnings
from copy import copy

import healpy as hp
import numpy as np

from rubin_scheduler.scheduler.surveys import BaseMarkovSurvey
from rubin_scheduler.scheduler.utils import ObservationArray, int_binned_stat, order_observations
from rubin_scheduler.utils import DEFAULT_NSIDE, _angular_separation, _hpid2_ra_dec, hp_grow_argsort


class GreedySurvey(BaseMarkovSurvey):
    """
    Select pointings in a greedy way using a Markov Decision Process.
    """

    def __init__(
        self,
        basis_functions,
        basis_weights,
        filtername="r",
        block_size=1,
        smoothing_kernel=None,
        nside=DEFAULT_NSIDE,
        dither=True,
        seed=42,
        ignore_obs=None,
        survey_name=None,
        scheduler_note=None,
        nexp=2,
        exptime=30.0,
        detailers=None,
        camera="LSST",
        area_required=None,
        fields=None,
        **kwargs,
    ):
        extra_features = {}

        self.filtername = filtername
        self.block_size = block_size
        self.nexp = nexp
        self.exptime = exptime

        super(GreedySurvey, self).__init__(
            basis_functions=basis_functions,
            basis_weights=basis_weights,
            extra_features=extra_features,
            smoothing_kernel=smoothing_kernel,
            ignore_obs=ignore_obs,
            nside=nside,
            survey_name=survey_name,
            scheduler_note=scheduler_note,
            dither=dither,
            seed=seed,
            detailers=detailers,
            camera=camera,
            area_required=area_required,
            fields=fields,
            **kwargs,
        )

    def _generate_survey_name(self):
        self.survey_name = f"Greedy {self.filtername}"

    def generate_observations_rough(self, conditions):
        """
        Just point at the highest reward healpix
        """
        self.reward = self.calc_reward_function(conditions)

        # Check if we need to spin the tesselation
        if self.dither & (conditions.night != self.night):
            self._spin_fields(conditions)
            self.night = copy(conditions.night)

        # Let's find the best N from the fields
        order = np.argsort(self.reward)[::-1]
        # Crop off any NaNs or Infs
        order = order[np.isfinite(self.reward[order])]

        iter = 0
        while True:
            best_hp = order[iter * self.block_size : (iter + 1) * self.block_size]
            best_fields = np.unique(self.hp2fields[best_hp])
            observations = []
            for field in best_fields:
                obs = ObservationArray()
                obs["RA"] = self.fields["RA"][field]
                obs["dec"] = self.fields["dec"][field]
                obs["rotSkyPos"] = 0.0
                obs["filter"] = self.filtername
                obs["nexp"] = self.nexp
                obs["exptime"] = self.exptime
                obs["scheduler_note"] = self.scheduler_note
                observations.append(obs)
                break
            iter += 1
            if len(observations) > 0 or (iter + 2) * self.block_size > len(order):
                break
        return observations


class BlobSurvey(GreedySurvey):
    """Select observations in large, mostly contiguous, blobs.

    Parameters
    ----------
    filtername1 : `str`
        The filter to observe in.
    filtername2 : `str`
        The filter to pair with the first observation. If set to None,
        no pair will be observed.
    slew_approx : `float`
        The approximate slewtime between neerby fields (seconds). Used
        to calculate how many observations can be taken in the
        desired time block.
    filter_change_approx : `float`
         The approximate time it takes to change filters (seconds).
    read_approx : `float`
        The approximate time required to readout the camera (seconds).
    exptime : `float`
        The total on-sky exposure time per visit.
    nexp : `int`
        The number of exposures to take in a visit.
    exp_dict : `dict`
        If set, should have keys of filtername and values of ints that
        are the nuber of exposures to take per visit. For estimating
        block time, nexp is still used.
    ideal_pair_time : `float`
        The ideal time gap wanted between observations to the same
        pointing (minutes)
    flush_time : `float`
        The time past the final expected exposure to flush the queue.
        Keeps observations from lingering past when they should be
        executed. (minutes)
    twilight_scale : `bool`
        Scale the block size to fill up to twilight. Set to False if
        running in twilight
    in_twilight : `bool`
        Scale the block size to stay within twilight time.
    check_scheduled : `bool`
        Check if there are scheduled observations and scale blob size
        to match
    min_area : `float`
        If set, demand the reward function have an area of so many
        square degrees before executing
    grow_blob : `bool`
        If True, try to grow the blob from the global maximum. Otherwise,
        just use a simple sort.  Simple sort will not constrain the
        blob to be contiguous.
    max_radius_peak : `float`
        The maximum radius to demand things be within the maximum of
        the reward function. (degrees) Note that traveling salesman
        solver can have rare failures if this is set too large (probably
        issue with projection effects or something).

    Notes
    -----
    The `scheduler_note` for the BlobSurvey will be set from the
    `survey_name`. A typical Detailer for the blob survey
    then adds onto this note to identify the first vs. second visit of
    the pair. Because the `scheduler_note` is modified, users do not set
    `scheduler_note` directly.
    """

    def __init__(
        self,
        basis_functions,
        basis_weights,
        filtername1="r",
        filtername2="g",
        slew_approx=7.5,
        filter_change_approx=140.0,
        read_approx=2.4,
        exptime=30.0,
        nexp=2,
        nexp_dict=None,
        ideal_pair_time=22.0,
        flush_time=30.0,
        smoothing_kernel=None,
        nside=DEFAULT_NSIDE,
        dither=True,
        seed=42,
        ignore_obs=None,
        survey_name=None,
        detailers=None,
        camera="LSST",
        twilight_scale=True,
        in_twilight=False,
        check_scheduled=True,
        min_area=None,
        grow_blob=True,
        area_required=None,
        max_radius_peak=40.0,
        fields=None,
        search_radius=None,
        alt_max=-9999,
        az_range=-9999,
        target_name=None,
        observation_reason=None,
        science_program=None,
    ):
        if search_radius is not None:
            warnings.warn("search_radius unused, remove kwarg", DeprecationWarning, 2)
        if alt_max != -9999:
            warnings.warn("alt_max unused, remove kwarg", DeprecationWarning, 2)
        if az_range != -9999:
            warnings.warn("az_range unused, remove kwarg", DeprecationWarning, 2)

        self.filtername1 = filtername1
        self.filtername2 = filtername2

        self.ideal_pair_time = ideal_pair_time

        if survey_name is None:
            self._generate_survey_name()
        else:
            self.survey_name = survey_name

        super(BlobSurvey, self).__init__(
            basis_functions=basis_functions,
            basis_weights=basis_weights,
            filtername=None,
            block_size=0,
            smoothing_kernel=smoothing_kernel,
            dither=dither,
            seed=seed,
            ignore_obs=ignore_obs,
            nside=nside,
            detailers=detailers,
            camera=camera,
            area_required=area_required,
            fields=fields,
            survey_name=self.survey_name,
            target_name=target_name,
            science_program=science_program,
            observation_reason=observation_reason,
        )
        self.flush_time = flush_time / 60.0 / 24.0  # convert to days
        self.nexp = nexp
        self.nexp_dict = nexp_dict
        self.exptime = exptime
        self.slew_approx = slew_approx
        self.read_approx = read_approx
        self.hpids = np.arange(hp.nside2npix(self.nside))
        self.twilight_scale = twilight_scale
        self.in_twilight = in_twilight
        self.grow_blob = grow_blob
        self.max_radius_peak = np.radians(max_radius_peak)

        if self.twilight_scale & self.in_twilight:
            warnings.warn("Both twilight_scale and in_twilight are set to True. That is probably wrong.")

        self.min_area = min_area
        self.check_scheduled = check_scheduled
        # If we are taking pairs in same filter, no need to add filter
        # change time.
        if filtername1 == filtername2:
            filter_change_approx = 0
        # Compute the minimum time needed to observe a blob (or observe,
        # then repeat.)
        if filtername2 is not None:
            self.time_needed = (
                (self.ideal_pair_time * 60.0 * 2.0 + self.exptime + self.read_approx + filter_change_approx)
                / 24.0
                / 3600.0
            )  # Days
        else:
            self.time_needed = (
                (self.ideal_pair_time * 60.0 + self.exptime + self.read_approx) / 24.0 / 3600.0
            )  # Days
        self.filter_set = set(filtername1)
        if filtername2 is None:
            self.filter2_set = self.filter_set
        else:
            self.filter2_set = set(filtername2)

        self.ra, self.dec = _hpid2_ra_dec(self.nside, self.hpids)

        self.counter = 1  # start at 1, because 0 is default in empty obs

        self.pixarea = hp.nside2pixarea(self.nside, degrees=True)

        # If we are only using one filter, this could be useful
        if (self.filtername2 is None) | (self.filtername1 == self.filtername2):
            self.filtername = self.filtername1

    def _generate_survey_name(self):
        self.survey_name = "Pairs"
        self.survey_name += f" {self.ideal_pair_time :.1f}"
        self.survey_name += f" {self.filtername1}"
        if self.filtername2 is None:
            self.survey_name += f"_{self.filtername1}"
        else:
            self.survey_name += f"_{self.filtername2}"

    def _check_feasibility(self, conditions):
        """
        Check if the survey is feasable in the current conditions.
        """
        for bf in self.basis_functions:
            result = bf.check_feasibility(conditions)
            if not result:
                return result

        # If we need to check that the reward function has enough
        # area available
        if self.min_area is not None:
            reward = 0
            for bf, weight in zip(self.basis_functions, self.basis_weights):
                basis_value = bf(conditions)
                reward += basis_value * weight
            # Are there any valid reward pixels remaining
            if np.sum(np.isfinite(reward)) > 0:
                max_reward_indx = np.min(np.where(reward == np.nanmax(reward)))
                distances = _angular_separation(
                    self.ra, self.dec, self.ra[max_reward_indx], self.dec[max_reward_indx]
                )
                valid_pix = np.where((np.isnan(reward) == False) & (distances < self.max_radius_peak))[0]
                if np.size(valid_pix) * self.pixarea < self.min_area:
                    result = False
            else:
                result = False
        return result

    def _set_block_size(self, conditions):
        """
        Update the block size if it's getting near a break point.
        """

        # If we are trying to get things done before twilight
        if self.twilight_scale:
            available_time = conditions.sun_n18_rising - conditions.mjd
            available_time *= 24.0 * 60.0  # to minutes
            n_ideal_blocks = available_time / self.ideal_pair_time
        else:
            n_ideal_blocks = 4

        # If we are trying to get things done before a scheduled simulation
        if self.check_scheduled:
            if len(conditions.scheduled_observations) > 0:
                available_time = np.min(conditions.scheduled_observations) - conditions.mjd
                available_time *= 24.0 * 60.0  # to minutes
                n_blocks = available_time / self.ideal_pair_time
                if n_blocks < n_ideal_blocks:
                    n_ideal_blocks = n_blocks

        # If we are trying to complete before twilight ends or
        # the night ends
        if self.in_twilight:
            at1 = conditions.sun_n12_rising - conditions.mjd
            at2 = conditions.sun_n18_setting - conditions.mjd
            times = np.array([at1, at2])
            times = times[np.where(times > 0)]
            available_time = np.min(times) if len(times) > 0 else 0.0
            available_time *= 24.0 * 60.0  # to minutes
            n_blocks = available_time / self.ideal_pair_time
            if n_blocks < n_ideal_blocks:
                n_ideal_blocks = n_blocks

        if n_ideal_blocks >= 3:
            self.nvisit_block = int(
                np.floor(
                    self.ideal_pair_time
                    * 60.0
                    / (self.slew_approx + self.exptime + self.read_approx * (self.nexp - 1))
                )
            )
        else:
            # Now we can stretch or contract the block size to
            # allocate the
            # remainder time until twilight starts
            # We can take the remaining time and try to do 1,2,
            # or 3 blocks.
            possible_times = available_time / np.arange(1, 4)
            diff = np.abs(self.ideal_pair_time - possible_times)
            best_block_time = np.max(possible_times[np.where(diff == np.min(diff))])
            self.nvisit_block = int(
                np.floor(
                    best_block_time
                    * 60.0
                    / (self.slew_approx + self.exptime + self.read_approx * (self.nexp - 1))
                )
            )

        # The floor can set block to zero, make it possible to to just one
        if self.nvisit_block <= 0:
            self.nvisit_block = 1

    def calc_reward_function(self, conditions):
        # Set the number of observations we are going to try and take
        self._set_block_size(conditions)
        #  Computing reward like usual with basis functions and weights
        if self._check_feasibility(conditions):
            self.reward = 0
            indx = np.arange(hp.nside2npix(self.nside))
            for bf, weight in zip(self.basis_functions, self.basis_weights):
                basis_value = bf(conditions, indx=indx)
                self.reward += basis_value * weight
            if self.smoothing_kernel is not None:
                self.smooth_reward()
        else:
            self.reward = -np.inf
            return self.reward

        if self.area_required is not None:
            max_indices = np.where(self.reward == np.nanmax(self.reward))[0]
            if np.size(max_indices) == 0:
                # This is the case if everything is masked
                self.reward = -np.inf
            else:
                max_reward_indx = np.min(max_indices)
                distances = _angular_separation(
                    self.ra,
                    self.dec,
                    self.ra[max_reward_indx],
                    self.dec[max_reward_indx],
                )
                good_area = np.where((np.abs(self.reward) >= 0) & (distances < self.max_radius_peak))[
                    0
                ].size * hp.nside2pixarea(self.nside)
                if good_area < self.area_required:
                    self.reward = -np.inf

        self.reward_checked = True
        return self.reward

    def simple_order_sort(self):
        """Fall back if we can't link contiguous blobs in the reward map"""

        # Assuming reward has already been calculated

        potential_hp = np.where(np.isfinite(self.reward))[0]

        # Note, using nanmax, so masked pixels might be included in
        # the pointing. I guess I should document that it's not
        # "NaN pixels can't be observed", but
        # "non-NaN pixels CAN be observed", which probably is
        # not intuitive.
        ufields, reward_by_field = int_binned_stat(
            self.hp2fields[potential_hp], self.reward[potential_hp], statistic=np.nanmax
        )
        # chop off any nans
        not_nans = np.where(np.isfinite(reward_by_field))
        ufields = ufields[not_nans]
        reward_by_field = reward_by_field[not_nans]

        order = np.argsort(reward_by_field)[::-1]
        ufields = ufields[order]
        self.best_fields = ufields[0 : self.nvisit_block]

    def generate_observations_rough(self, conditions):
        """
        Find a good block of observations.
        """

        self.reward = self.calc_reward_function(conditions)

        # Mask off pixels that are far away from the maximum.
        max_reward_indx = np.min(np.where(self.reward == np.nanmax(self.reward)))
        distances = _angular_separation(
            self.ra, self.dec, self.ra[max_reward_indx], self.dec[max_reward_indx]
        )

        self.reward[np.where(distances > self.max_radius_peak)] = np.nan
        # Check if we need to spin the tesselation
        if self.dither & (conditions.night != self.night):
            self._spin_fields(conditions)
            self.night = copy(conditions.night)

        if self.grow_blob:
            # Note, returns highest first
            ordered_hp = hp_grow_argsort(self.reward)
            ordered_fields = self.hp2fields[ordered_hp]
            orig_order = np.arange(ordered_fields.size)
            # Remove duplicate field pointings
            _u_of, u_indx = np.unique(ordered_fields, return_index=True)
            new_order = np.argsort(orig_order[u_indx])
            best_fields = ordered_fields[u_indx[new_order]]

            if np.size(best_fields) < self.nvisit_block:
                # Let's fall back to the simple sort
                self.simple_order_sort()
            else:
                self.best_fields = best_fields[0 : self.nvisit_block]
        else:
            self.simple_order_sort()

        if len(self.best_fields) == 0:
            # everything was nans, or self.nvisit_block was zero
            return []

        better_order = order_observations(
            self.fields["RA"][self.best_fields], self.fields["dec"][self.best_fields]
        )

        # XXX-TODO: Could try to roll better_order to start at
        # the nearest/fastest slew from current position.
        observations = []
        counter2 = 0
        flush_time = conditions.mjd + self.time_needed + self.flush_time

        for i, indx in enumerate(better_order):
            field = self.best_fields[indx]
            obs = ObservationArray()
            obs["RA"] = self.fields["RA"][field]
            obs["dec"] = self.fields["dec"][field]
            obs["rotSkyPos"] = 0.0
            obs["filter"] = self.filtername1
            if self.nexp_dict is None:
                obs["nexp"] = self.nexp
            else:
                obs["nexp"] = self.nexp_dict[self.filtername1]
            obs["exptime"] = self.exptime
            obs["scheduler_note"] = self.scheduler_note
            obs["block_id"] = self.counter
            obs["flush_by_mjd"] = flush_time
            observations.append(obs)
            counter2 += 1

        result = observations
        return result



__all__ = ("BaseSurvey", "BaseMarkovSurvey")

from copy import copy, deepcopy
from functools import cached_property

import healpy as hp
import numpy as np
import pandas as pd

from rubin_scheduler.scheduler.detailers import TrackingInfoDetailer, ZeroRotDetailer
from rubin_scheduler.scheduler.utils import (
    HpInLsstFov,
    ObservationArray,
    comcam_tessellate,
    thetaphi2xyz,
    xyz2thetaphi,
)
from rubin_scheduler.site_models import _read_fields
from rubin_scheduler.utils import DEFAULT_NSIDE, _build_tree, _hpid2_ra_dec, _xyz_from_ra_dec


class BaseSurvey:
    """A baseclass for survey objects.

    Parameters
    ----------
    basis_functions : list
        List of basis_function objects
    extra_features : list XXX--should this be a dict for clarity?
        List of any additional features the survey may want to use
        e.g., for computing final dither positions.
    extra_basis_functions : dict of rubin_scheduler.scheduler.basis_function
        Extra basis function objects. Typically not passed in, but et
        in the __init__.
    ignore_obs : list of str (None)
        If an incoming observation has this string in the note, ignore it.
        Handy if one wants to ignore DD fields or observations
        requested by self. Take note, if a survey is called 'mysurvey23',
        setting ignore_obs to 'mysurvey2' will ignore it because
        'mysurvey2' is a substring of 'mysurvey23'.
    detailers : list of rubin_scheduler.scheduler.detailers objects
        The detailers to apply to the list of observations.
    scheduled_obs : np.array
        An array of MJD values for when observations should execute.
    target_name : `str`
        A target name label. Will be added to a final detailer, so should
        override any target_name set elsewhere. Default None.
    science_program : `str`
        A science program label.  Will be added to a final detailer, so should
        override any science_program set elsewhere. Default None.
    observation_reason : `str`
        An observation reason label.  Will be added to a final detailer, so
        should  override any observation_reason set elsewhere. Default None.
    """

    def __init__(
        self,
        basis_functions,
        extra_features=None,
        extra_basis_functions=None,
        ignore_obs=None,
        survey_name=None,
        scheduler_note=None,
        nside=DEFAULT_NSIDE,
        detailers=None,
        scheduled_obs=None,
        target_name=None,
        science_program=None,
        observation_reason=None,
    ):
        if ignore_obs is None:
            ignore_obs = []

        if isinstance(ignore_obs, str):
            ignore_obs = [ignore_obs]

        self.nside = nside
        if survey_name is None:
            self._generate_survey_name()
        else:
            self.survey_name = survey_name
        self.scheduler_note = scheduler_note
        if self.scheduler_note is None:
            self.scheduler_note = survey_name

        self.ignore_obs = ignore_obs

        self.reward = None
        self.survey_index = None

        self.basis_functions = basis_functions

        if extra_features is None:
            self.extra_features = {}
        else:
            self.extra_features = extra_features
        if extra_basis_functions is None:
            self.extra_basis_functions = {}
        else:
            self.extra_basis_functions = extra_basis_functions

        self.reward_checked = False

        # Attribute to track if the reward function is up-to-date.
        self.reward_checked = False

        # If there's no detailers, add one to set rotation to near zero
        if detailers is None:
            self.detailers = [ZeroRotDetailer(nside=nside)]
        else:
            # If there are detailers -- copy them, because we're
            # about to change the list (and users can be reusing this list).
            self.detailers = deepcopy(detailers)

        # Scheduled observations
        self.scheduled_obs = scheduled_obs

        # Strings to pass onto observations
        if (target_name is not None) | (science_program is not None) | (observation_reason is not None):
            self.detailers.append(
                TrackingInfoDetailer(
                    target_name=target_name,
                    science_program=science_program,
                    observation_reason=observation_reason,
                )
            )

    @cached_property
    def roi_hpid(self):
        return None

    @cached_property
    def roi_mask(self):
        if self.roi_hpid is None:
            mask = np.ones(hp.nside2npix(self.nside), dtype=bool)
        else:
            mask = np.zeros(hp.nside2npix(self.nside), dtype=bool)
            mask[self.roi_hpid] = True

        return mask

    def _generate_survey_name(self):
        self.survey_name = ""

    def get_scheduled_obs(self):
        return self.scheduled_obs

    def add_observations_array(self, observations_array_in, observations_hpid_in):
        """Add an array of observations rather than one at a time

        Parameters
        ----------
        observations_array_in : ObservationArray
            An array of completed observations,
            rubin_scheduler.scheduler.utils.ObservationArray
        observations_hpid_in : np.array
            Same as observations_array_in, but larger and with an
            additional column for HEALpix id. Each observation is
            listed mulitple times, once for every HEALpix it overlaps.
        """

        # Just to be sure things are sorted
        observations_array_in.sort(order="mjd")
        observations_hpid_in.sort(order="mjd")

        # Copy so we don't prune things for other survey objects
        observations_array = observations_array_in.copy()
        observations_hpid = observations_hpid_in.copy()

        for ig in self.ignore_obs:
            not_ignore = np.where(np.char.find(observations_array["scheduler_note"], ig) == -1)[0]
            observations_array = observations_array[not_ignore]

            not_ignore = np.where(np.char.find(observations_hpid["scheduler_note"], ig) == -1)[0]
            observations_hpid = observations_hpid[not_ignore]

        for feature in self.extra_features:
            self.extra_features[feature].add_observations_array(observations_array, observations_hpid)
        for bf in self.extra_basis_functions:
            self.extra_basis_functions[bf].add_observations_array(observations_array, observations_hpid)
        for bf in self.basis_functions:
            bf.add_observations_array(observations_array, observations_hpid)
        for detailer in self.detailers:
            detailer.add_observations_array(observations_array, observations_hpid)
        self.reward_checked = False

    def add_observation(self, observation, **kwargs):
        # Check each posible ignore string
        checks = [io not in str(observation["scheduler_note"]) for io in self.ignore_obs]
        # ugh, I think here I have to assume observation is an
        # array and not a dict.
        if all(checks):
            for feature in self.extra_features:
                self.extra_features[feature].add_observation(observation, **kwargs)
            for bf in self.extra_basis_functions:
                self.extra_basis_functions[bf].add_observation(observation, **kwargs)
            for bf in self.basis_functions:
                bf.add_observation(observation, **kwargs)
            for detailer in self.detailers:
                detailer.add_observation(observation, **kwargs)
            self.reward_checked = False

    def _check_feasibility(self, conditions):
        """
        Check if the survey is feasible in the current conditions

        Returns
        -------
        result : `bool`
        """
        result = True
        for bf in self.basis_functions:
            result = bf.check_feasibility(conditions)
            if not result:
                return result
        return result

    def calc_reward_function(self, conditions):
        """
        Parameters
        ----------
        conditions : rubin_scheduler.scheduler.features.Conditions object

        Returns
        -------
        reward : float (or array)

        """
        if self._check_feasibility(conditions):
            self.reward = 0
        else:
            # If we don't pass feasability
            self.reward = -np.inf

        self.reward_checked = True
        return self.reward

    def generate_observations_rough(self, conditions):
        """
        Returns
        -------
        one of:
            1) None
            2) A list of observations
        """
        # If the reward function hasn't been updated with the
        # latest info, calculate it
        if not self.reward_checked:
            self.reward = self.calc_reward_function(conditions)
        obs = ObservationArray()
        return [obs]

    def generate_observations(self, conditions):
        observations = self.generate_observations_rough(conditions)
        if np.size(observations) > 0:
            for detailer in self.detailers:
                observations = detailer(observations, conditions)
        return observations

    def viz_config(self):
        # This primarily lives in schedview but could be somewhat added here
        pass

    def __repr__(self):
        try:
            repr = f"<{self.__class__.__name__} survey_name='{self.survey_name}' at {hex(id(self))}>"
        except AttributeError:
            repr = f"<{self.__class__.__name__} at {hex(id(self))}>"

        return repr

    def _reward_to_scalars(self, reward):
        if np.isscalar(reward):
            scalar_reward = reward
            if np.isnan(reward) or reward == -np.inf:
                unmasked_area = 0
                scalar_reward = -np.inf
            else:
                try:
                    pix_area = hp.nside2pixarea(self.nside, degrees=True)
                    unmasked_area = pix_area * np.count_nonzero(self.roi_mask)
                except AttributeError:
                    unmasked_area = 4 * np.pi * (180 / np.pi) ** 2
        else:
            reward_in_roi = np.where(self.roi_mask, reward, -np.inf)
            pix_area = hp.nside2pixarea(self.nside, degrees=True)
            unmasked_area = pix_area * np.count_nonzero(reward_in_roi > -np.inf)
            if unmasked_area == 0:
                scalar_reward = -np.inf
            else:
                scalar_reward = np.nanmax(reward_in_roi)

        return scalar_reward, unmasked_area

    def make_reward_df(self, conditions, accum=True):
        """Create a pandas.DataFrame describing the reward from the survey.

        Parameters
        ----------
        conditions : `rubin_scheduler.scheduler.features.Conditions`
            Conditions for which rewards are to be returned
        accum : `bool`
            Include accumulated reward (more compute intensive)
            Defaults to True

        Returns
        -------
        reward_df : `pandas.DataFrame`
            A table of surveys listing the rewards.
        """

        feasibility = []
        max_rewards = []
        basis_areas = []
        accum_rewards = []
        accum_areas = []
        bf_label = []
        bf_class = []
        basis_functions = []
        basis_weights = []

        try:
            full_basis_weights = self.basis_weights
        except AttributeError:
            full_basis_weights = [1.0 for df in self.basis_functions]

        short_labels = self.bf_short_labels

        _, scalar_area = self._reward_to_scalars(1)

        for weight, basis_function in zip(full_basis_weights, self.basis_functions):
            bf_label.append(short_labels[basis_function.label()])
            bf_class.append(basis_function.__class__.__name__)
            bf_reward = basis_function(conditions)
            max_reward, basis_area = self._reward_to_scalars(bf_reward)

            if basis_area == 0:
                this_feasibility = False
            else:
                this_feasibility = np.array(basis_function.check_feasibility(conditions)).any()

            feasibility.append(this_feasibility)
            max_rewards.append(max_reward)
            basis_areas.append(basis_area)

            if accum:
                basis_functions.append(basis_function)
                basis_weights.append(weight)
                test_survey = deepcopy(self)
                test_survey.basis_functions = basis_functions
                test_survey.basis_weights = basis_weights
                this_accum_reward = test_survey.calc_reward_function(conditions)
                accum_reward, accum_area = self._reward_to_scalars(this_accum_reward)
                accum_rewards.append(accum_reward)
                accum_areas.append(accum_area)

        reward_data = {
            "basis_function": bf_label,
            "basis_function_class": bf_class,
            "feasible": feasibility,
            "max_basis_reward": max_rewards,
            "basis_area": basis_areas,
            "basis_weight": full_basis_weights,
        }
        if accum:
            reward_data["max_accum_reward"] = accum_rewards
            reward_data["accum_area"] = accum_areas

        reward_df = pd.DataFrame(reward_data)

        return reward_df

    def reward_changes(self, conditions):
        """List the rewards for each basis function used by the survey.

        Parameters
        ----------
        conditions : `rubin_scheduler.scheduler.features.Conditions`
            Conditions for which rewards are to be returned

        Returns
        -------
        rewards : `list`
            A list of tuples, each with a basis function name and the
            maximum reward returned by that basis function for the
            provided conditions.
        """

        reward_values = []
        basis_functions = []
        basis_weights = []

        try:
            full_basis_weights = self.basis_weights
        except AttributeError:
            full_basis_weights = [1 for bf in self.basis_functions]

        for weight, basis_function in zip(full_basis_weights, self.basis_functions):
            test_survey = deepcopy(self)
            basis_functions.append(basis_function)
            test_survey.basis_functions = basis_functions
            basis_weights.append(weight)
            test_survey.basis_weights = basis_weights
            try:
                reward_values.append(np.nanmax(test_survey.calc_reward_function(conditions)))
            except IndexError:
                reward_values.append(None)

        bf_names = [bf.label() for bf in self.basis_functions]
        return list(zip(bf_names, reward_values))

    @property
    def bf_short_labels(self):
        try:
            long_labels = [bf.label() for bf in self.basis_functions]
        except AttributeError:
            return []

        label_bases = [label.split(" @")[0] for label in long_labels]
        duplicated_labels = set([label for label in label_bases if label_bases.count(label) > 1])
        short_labels = []
        label_count = {k: 0 for k in duplicated_labels}
        for label_base in label_bases:
            if label_base in duplicated_labels:
                label_count[label_base] += 1
                short_labels.append(f"{label_base} {label_count[label_base]}")
            else:
                short_labels.append(label_base)

        label_map = dict(zip(long_labels, short_labels))

        return label_map


def rotx(theta, x, y, z):
    """rotate the x,y,z points theta radians about x axis"""
    sin_t = np.sin(theta)
    cos_t = np.cos(theta)
    xp = x
    yp = y * cos_t + z * sin_t
    zp = -y * sin_t + z * cos_t
    return xp, yp, zp


class BaseMarkovSurvey(BaseSurvey):
    """A Markov Decision Function survey object. Uses Basis functions
    to compute a final reward function and decide what to observe based
    on the reward. Includes methods for dithering and defaults to
    dithering nightly.

    Parameters
    ----------
    basis_function : list of rubin_scheduler.scheduler.basis_function

    basis_weights : list of float
        Must be same length as basis_function
    seed : hashable
        Random number seed, used for randomly orienting sky tessellation.
    camera : str ('LSST')
        Should be 'LSST' or 'comcam'
    fields : np.array (None)
        An array of field positions. Should be numpy array with columns
        of "RA" and "dec" in radians. If none,
        site_models.read_fields or utils.comcam_tessellate is used to
        read field positions.
    area_required : float (None)
        The valid area that should be present in the reward
        function (square degrees).
    npositions : int (7305)
        The number of dither positions to pre-compute. Defaults
        to 7305 (so good for 20 years)
    """

    def __init__(
        self,
        basis_functions,
        basis_weights,
        extra_features=None,
        smoothing_kernel=None,
        ignore_obs=None,
        survey_name=None,
        scheduler_note=None,
        nside=DEFAULT_NSIDE,
        seed=42,
        dither=True,
        detailers=None,
        camera="LSST",
        fields=None,
        area_required=None,
        npositions=7305,
        target_name=None,
        science_program=None,
        observation_reason=None,
    ):
        super(BaseMarkovSurvey, self).__init__(
            basis_functions=basis_functions,
            extra_features=extra_features,
            ignore_obs=ignore_obs,
            survey_name=survey_name,
            scheduler_note=scheduler_note,
            nside=nside,
            detailers=detailers,
            target_name=target_name,
            science_program=science_program,
            observation_reason=observation_reason,
        )

        self.basis_weights = basis_weights
        # Check that weights and basis functions are same length
        if len(basis_functions) != np.size(basis_weights):
            raise ValueError("basis_functions and basis_weights must be same length.")

        self.camera = camera
        # Load the OpSim field tesselation and map healpix to fields
        if fields is None:
            if self.camera == "LSST":
                ra, dec = _read_fields()
                self.fields_init = np.empty(ra.size, dtype=list(zip(["RA", "dec"], [float, float])))
                self.fields_init["RA"] = ra
                self.fields_init["dec"] = dec
            elif self.camera == "comcam":
                self.fields_init = comcam_tessellate()
            else:
                ValueError('camera %s unknown, should be "LSST" or "comcam"' % camera)
        else:
            self.fields_init = fields
        self.fields = self.fields_init.copy()
        self.hp2fields = np.array([])
        self._hp2fieldsetup(self.fields["RA"], self.fields["dec"])

        if smoothing_kernel is not None:
            self.smoothing_kernel = np.radians(smoothing_kernel)
        else:
            self.smoothing_kernel = None

        if area_required is None:
            self.area_required = area_required
        else:
            self.area_required = area_required * (np.pi / 180.0) ** 2  # To steradians

        # Start tracking the night
        self.night = -1

        self.dither = dither

        # Generate and store rotation positions to use.
        # This way, if different survey objects are seeded the same,
        # they will use the same dither positions each night
        rng = np.random.default_rng(seed)
        self.lon = rng.random(npositions) * np.pi * 2
        # Make sure latitude points spread correctly
        # http://mathworld.wolfram.com/SpherePointPicking.html
        self.lat = np.arccos(2.0 * rng.random(npositions) - 1.0)
        self.lon2 = rng.random(npositions) * np.pi * 2

    def _check_feasibility(self, conditions):
        """
        Check if the survey is feasable in the current conditions
        """
        for bf in self.basis_functions:
            result = bf.check_feasibility(conditions)
            if not result:
                return result
        if self.area_required is not None:
            reward = self.calc_reward_function(conditions)
            good_pix = np.where(np.isfinite(reward) == True)[0]
            area = hp.nside2pixarea(self.nside) * np.size(good_pix)
            if area < self.area_required:
                return False
        return result

    def _hp2fieldsetup(self, ra, dec):
        """Map each healpixel to nearest field. This will only work
        if healpix resolution is higher than field resolution.

        Parameters
        ----------
        ra : `float`
            The RA of the possible pointings (radians)
        dec : `float`
            The decs of the possible pointings (radians)
        """
        self.hp2fields = np.zeros(hp.nside2npix(self.nside), dtype=int)
        if self.camera == "LSST":
            pointing2hpindx = HpInLsstFov(nside=self.nside)
            for i in range(len(ra)):
                hpindx = pointing2hpindx(ra[i], dec[i], rotSkyPos=0.0)
                self.hp2fields[hpindx] = i
        elif self.camera == "comcam":
            # Let's just map each healpix to the closest field location
            tree = _build_tree(ra, dec)
            hp_ra, hp_dec = _hpid2_ra_dec(self.nside, np.arange(hp.nside2npix(self.nside)))
            x, y, z = _xyz_from_ra_dec(hp_ra, hp_dec)
            dist, ind = tree.query(np.vstack([x, y, z]).T, k=1)
            self.hp2fields = ind

    def _spin_fields(self, conditions, lon=None, lat=None, lon2=None):
        """Spin the field tessellation to generate a random orientation

        The default field tesselation is rotated randomly in longitude,
        and then the pole is rotated to a random point on the sphere.

        Parameters
        ----------
        lon : float (None)
            The amount to initially rotate in longitude (radians).
            Will use a random value
            between 0 and 2 pi if None (default).
        lat : float (None)
            The amount to rotate in latitude (radians).
        lon2 : float (None)
            The amount to rotate the pole in longitude (radians).
        """
        if lon is None:
            lon = self.lon[conditions.night]
        if lat is None:
            lat = self.lat[conditions.night]
        if lon2 is None:
            lon2 = self.lon2[conditions.night]

        # rotate longitude
        ra = (self.fields_init["RA"] + lon) % (2.0 * np.pi)
        dec = copy(self.fields_init["dec"])

        # Now to rotate ra and dec about the x-axis
        x, y, z = thetaphi2xyz(ra, dec + np.pi / 2.0)
        xp, yp, zp = rotx(lat, x, y, z)
        theta, phi = xyz2thetaphi(xp, yp, zp)
        dec = phi - np.pi / 2
        ra = theta + np.pi

        # One more RA rotation
        ra = (ra + lon2) % (2.0 * np.pi)

        self.fields["RA"] = ra
        self.fields["dec"] = dec
        # Rebuild the kdtree with the new positions
        # XXX-may be doing some ra,dec to conversions xyz more
        # than needed.
        self._hp2fieldsetup(ra, dec)

    def smooth_reward(self):
        """If we want to smooth the reward function."""
        if hp.isnpixok(self.reward.size):
            # Need to swap NaNs to hp.UNSEEN so smoothing doesn't
            # spread mask
            reward_temp = copy(self.reward)
            mask = np.isnan(reward_temp)
            reward_temp[mask] = hp.UNSEEN
            self.reward_smooth = hp.sphtfunc.smoothing(reward_temp, fwhm=self.smoothing_kernel, verbose=False)
            self.reward_smooth[mask] = np.nan
            self.reward = self.reward_smooth

    def calc_reward_function(self, conditions):
        self.reward_checked = True
        if self._check_feasibility(conditions):
            self.reward = 0
            indx = np.arange(hp.nside2npix(self.nside))
            for bf, weight in zip(self.basis_functions, self.basis_weights):
                basis_value = bf(conditions, indx=indx)
                self.reward += basis_value * weight
        else:
            # If not feasable, negative infinity reward
            self.reward = -np.inf
            return self.reward
        if self.smoothing_kernel is not None:
            self.smooth_reward()

        if self.area_required is not None:
            good_area = np.where(np.abs(self.reward) >= 0)[0].size * hp.nside2pixarea(self.nside)
            if good_area < self.area_required:
                self.reward = -np.inf

        return self.reward

    def generate_observations_rough(self, conditions):
        self.reward = self.calc_reward_function(conditions)

        # Check if we need to spin the tesselation
        if self.dither & (conditions.night != self.night):
            self._spin_fields(conditions)
            self.night = copy(conditions.night)

        # XXX Use self.reward to decide what to observe.
        return None



__all__ = ("ScriptedSurvey",)

import logging

import healpy as hp
import numpy as np

from rubin_scheduler.scheduler.surveys import BaseSurvey
from rubin_scheduler.scheduler.utils import ObservationArray, ScheduledObservationArray
from rubin_scheduler.utils import DEFAULT_NSIDE, _angular_separation, _approx_ra_dec2_alt_az

log = logging.getLogger(__name__)


class ScriptedSurvey(BaseSurvey):
    """
    Take a set of scheduled observations and serve them up.

    Parameters
    ----------
    basis_functions : list of rubin_scheduler.scheduler.BasisFunction
        Basis functions to use. These are only used for computing
        survey feasibility. They do not contribute to the logic of
        how observations are selected. Basis functions that return
        HEALpix maps are ignored. Spatial masking is instead done
        with the `_check_alts_ha` method.
    id_start : `int` (1)
        The integer to start the "scripted id" field with. Bad things
        could happen if you have multiple scripted survey objects with
        the same scripted IDs.
    return_n_limit : `int` (10)
        The maximum number of observations to return. Set to high and
        your block of scheduled observations can run into twilight time.
    before_twi_check : `bool`
        Check if the returned observations have enough time to complete
        before twilight starts. (default True)
    filter_change_time : `float`
        The time needed to change filters. Default 120 seconds. Only
        used if before_twi_check is True.
    """

    def __init__(
        self,
        basis_functions,
        basis_weights=None,
        reward=1e6,
        ignore_obs=None,
        nside=DEFAULT_NSIDE,
        detailers=None,
        id_start=1,
        return_n_limit=10,
        survey_name=None,
        before_twi_check=True,
        filter_change_time=120,
    ):
        """"""
        self.extra_features = {}
        self.nside = nside
        self.reward_val = reward
        self.reward = -np.inf
        self.id_start = id_start
        self.return_n_limit = return_n_limit
        self.filter_change_time = filter_change_time / 3600 / 24.0  # to days
        if basis_weights is None:
            self.basis_weights = np.zeros(len(basis_functions))
        else:
            if np.max(np.abs(basis_weights)) > 0:
                raise ValueError("Basis function weights should be zero for ScriptedSurvey objects.")
            if len(basis_weights) != len(basis_functions):
                raise ValueError("Length of Basis function weights should match length of basis_functions.")
            self.basis_weights = basis_weights
        self.before_twi_check = before_twi_check
        # Attribute to cache results into
        self.observations = None
        # Track when the last call to generate_observations_rough was.
        self.last_mjd = -1
        super(ScriptedSurvey, self).__init__(
            basis_functions=basis_functions,
            ignore_obs=ignore_obs,
            nside=nside,
            detailers=detailers,
            survey_name=survey_name,
        )
        self.clear_script()

        # Just to be clear that the script should
        # be setting scheduler_note, not the survey.
        self.scheduler_note = None

    def add_observations_array(self, observations_array_in, observations_hpid_in):
        if self.obs_wanted is not None:
            # toss out things that should be ignored
            to_ignore = np.in1d(observations_array_in["scheduler_note"], self.ignore_obs)
            observations_array = observations_array_in[~to_ignore]

            good = np.in1d(observations_hpid_in["ID"], observations_array["ID"])
            observations_hpid = observations_hpid_in[good]

            for feature in self.extra_features:
                self.extra_features[feature].add_observations_array(observations_array, observations_hpid)
            for bf in self.extra_basis_functions:
                self.extra_basis_functions[bf].add_observations_array(observations_array, observations_hpid)
            for bf in self.basis_functions:
                bf.add_observations_array(observations_array, observations_hpid)
            for detailer in self.detailers:
                detailer.add_observations_array(observations_array, observations_hpid)

            if (self.obs_wanted is not None) & (np.size(self.obs_wanted) > 0):
                full_note_in = np.char.add(
                    observations_array_in["scheduler_note"], observations_array_in["filter"]
                )
                full_note_queue = np.char.add(self.obs_wanted["scheduler_note"], self.obs_wanted["filter"])

                indx = np.in1d(full_note_queue, full_note_in)

                self.obs_wanted["observed"][indx] = True
                self.scheduled_obs = self.obs_wanted["mjd"][~self.obs_wanted["observed"]]

    def add_observation(self, observation, indx=None, **kwargs):
        """Check if observation matches a scripted observation"""
        if (self.obs_wanted is not None) & (np.size(self.obs_wanted) > 0):
            # From base class
            checks = [io not in str(observation["scheduler_note"]) for io in self.ignore_obs]
            if all(checks):
                for feature in self.extra_features:
                    self.extra_features[feature].add_observation(observation, **kwargs)
                for bf in self.basis_functions:
                    bf.add_observation(observation, **kwargs)
                for detailer in self.detailers:
                    detailer.add_observation(observation, **kwargs)
                self.reward_checked = False

                key = observation["scheduler_note"][0] + observation["filter"][0]
                try:
                    # Could add an additional check here for if the observation
                    # is in the desired mjd window.
                    self.obs_wanted["observed"][self.note_filter_dict[key]] = True
                # If the key does not exist, nothing to do
                except KeyError:
                    return

    def calc_reward_function(self, conditions):
        """If there is an observation ready to go, execute it,
        otherwise, -inf"""
        observation = self.generate_observations_rough(conditions)
        if (observation is None) | (np.size(observation) == 0):
            self.reward = -np.inf
        else:
            self.reward = self.reward_val
        return self.reward

    def _slice2obs(self, obs_row):
        """take a slice and return a full observation object"""
        observation = ObservationArray()
        for key in [
            "RA",
            "dec",
            "filter",
            "exptime",
            "nexp",
            "scheduler_note",
            "target_name",
            "rotSkyPos",
            "rotTelPos",
            "flush_by_mjd",
        ]:
            observation[key] = obs_row[key]
        return observation

    def _check_alts_ha(self, observation, conditions):
        """Given scheduled observations, check which ones can be
        done in current conditions.

        Parameters
        ----------
        observation : np.array
            An array of scheduled observations. Probably generated with
            rubin_scheduler.scheduler.utils.scheduled_observation
        """
        # distance to the moon
        d_to_moon = _angular_separation(
            observation["RA"], observation["dec"], conditions.moon_ra, conditions.moon_dec
        )

        # Just do a fast ra,dec to alt,az conversion.
        alt, az = _approx_ra_dec2_alt_az(
            observation["RA"],
            observation["dec"],
            conditions.site.latitude_rad,
            None,
            conditions.mjd,
            lmst=conditions.lmst,
        )
        HA = conditions.lmst - observation["RA"] * 12.0 / np.pi
        HA[np.where(HA > 24)] -= 24
        HA[np.where(HA < 0)] += 24
        in_range = np.where(
            (d_to_moon > observation["moon_min_distance"])
            & (alt < observation["alt_max"])
            & (alt > observation["alt_min"])
            & ((HA > observation["HA_max"]) | (HA < observation["HA_min"]))
            & (conditions.sun_alt < observation["sun_alt_max"])
        )[0]

        # Also check the alt,az limits given by the conditions object
        count = in_range * 0
        if conditions.sky_alt_limits is not None:
            for limits in conditions.sky_alt_limits:
                ir = np.where((alt[in_range] >= np.min(limits)) & (alt[in_range] <= np.max(limits)))[0]
                count[ir] += 1
            good = np.where(count > 0)[0]
            in_range = in_range[good]
        # Check against kinematic limits too
        ir = np.where(
            (alt[in_range] >= np.min(conditions.tel_alt_limits))
            & (alt[in_range] <= np.max(conditions.tel_alt_limits))
        )[0]
        in_range = in_range[ir]

        count = in_range * 0
        if conditions.sky_az_limits is not None:
            for limits in conditions.sky_az_limits:
                az_min = limits[0]
                az_max = limits[1]
                if np.abs(az_max - az_min) >= (2 * np.pi):
                    count += 1
                else:
                    az_range = (az_max - az_min) % (2 * np.pi)
                    ir = np.where((az[in_range] - az_min) % (2 * np.pi) <= az_range)[0]
                    count[ir] += 1
            good = np.where(count > 0)[0]
            in_range = in_range[good]
        if np.abs(conditions.tel_az_limits[1] - conditions.tel_az_limits[0]) < (2 * np.pi):
            az_range = (conditions.tel_az_limits[1] - conditions.tel_az_limits[0]) % (2 * np.pi)
            ir = np.where((az[in_range] - conditions.tel_az_limits[0]) % (2 * np.pi) <= az_range)[0]
            in_range = in_range[ir]

        # Check that filter needed is mounted
        good = np.isin(observation["filter"][in_range], conditions.mounted_filters)
        in_range = in_range[good]

        return in_range

    def _check_list(self, conditions):
        """Check to see if the current mjd is good"""
        observations = None
        if self.obs_wanted.size > 0:
            # Scheduled observations that are in the right time
            # window and have not been executed
            in_time_window = np.where(
                (self.mjd_start < conditions.mjd)
                & (self.obs_wanted["flush_by_mjd"] > conditions.mjd)
                & (~self.obs_wanted["observed"])
            )[0]

            if np.size(in_time_window) > 0:
                pass_checks = self._check_alts_ha(self.obs_wanted[in_time_window], conditions)
                matches = in_time_window[pass_checks]

                # Also check that the filters are mounted
                match2 = np.isin(self.obs_wanted["filter"][matches], conditions.mounted_filters)
                matches = matches[match2]

            else:
                matches = []

            if np.size(matches) > 0:
                # Do not return too many observations
                if np.size(matches) > self.return_n_limit:
                    matches = matches[0 : self.return_n_limit]
                observations = self.obs_wanted[matches]
                # Need to check that none of these are masked by basis
                # functions
                reward = 0
                for bf, weight in zip(self.basis_functions, self.basis_weights):
                    basis_value = bf(conditions)
                    reward += basis_value * weight
                # If reward is an array, then it's a HEALpy map and we
                # need to interpolate to the actual positions we want.
                # now to interpolate to the reward positions
                if np.size(reward) > 1:
                    reward_interp = hp.get_interp_val(
                        reward,
                        np.degrees(observations["RA"]),
                        np.degrees(observations["dec"]),
                        lonlat=True,
                    )
                    valid_reward = np.isfinite(reward_interp)
                    observations = observations[valid_reward]

        return observations

    def clear_script(self):
        """set an empty list to serve up"""
        self.obs_wanted = ScheduledObservationArray(n=0)
        self.mjd_start = None
        self.scheduled_obs = None
        self.note_filter_dict = {}

    def set_script(self, obs_wanted, append=True, add_index=True):
        """
        Parameters
        ----------
        obs_wanted : np.array
            The observations that should be executed. Needs to have
            columns with dtype names:
            Should be from lsst.sim.scheduler.utils.scheduled_observation
        mjds : np.array
            The MJDs for the observaitons, should be same length as
            obs_list
        mjd_tol : float (15.)
            The tolerance to consider an observation as still good to
            observe (min)
        append : `bool`
            Should the obs_wanted be appended to any existing scheduled
            observations. Default True.
        add_index : `bool`
            Should the scheduler_note be modified to include a unique
            index value. Default True.
        """

        obs_wanted.sort(order=["mjd", "filter"])

        if add_index:
            sep = [", "] * obs_wanted.size
            indx = np.arange(self.id_start, self.id_start + obs_wanted.size, 1).astype(str)
            new_note = np.char.add(obs_wanted["scheduler_note"], sep)
            obs_wanted["scheduler_note"] = np.char.add(new_note, indx)
            self.id_start += obs_wanted.size
        else:
            # go through and pull the index as an int
            self.script_id_array = np.array(
                [int(note.split(", ")[-1]) for note in self.obs_wanted["scheduler_note"]]
            )
            if np.size(self.script_id_array) > 0:
                self.id_start = self.script_id_array.max() + 1

        if append & (self.obs_wanted is not None):
            self.obs_wanted = np.concatenate([self.obs_wanted, obs_wanted])
            self.obs_wanted.sort(order=["mjd", "filter"])
        else:
            self.obs_wanted = obs_wanted

        # check that we have valid unique keys for observations
        potential_keys = np.char.add(self.obs_wanted["scheduler_note"], self.obs_wanted["filter"])
        if np.size(np.unique(potential_keys)) < np.size(self.obs_wanted):
            msg = (
                "Scripted observations do not have unique scheduler_note "
                "+ filter values. Consider setting add_index=True"
            )
            raise ValueError(msg)

        self.mjd_start = self.obs_wanted["mjd"] - self.obs_wanted["mjd_tol"]
        # Here is the attribute that core scheduler checks to
        # broadcast scheduled observations in the conditions object.
        self.scheduled_obs = self.obs_wanted["mjd"]

        # Generate a dict so it can be fast to check if an observation matches
        # the combination of scheduler_note and filter
        self.note_filter_dict = {}
        for i, obs in enumerate(self.obs_wanted):
            self.note_filter_dict[obs["scheduler_note"] + obs["filter"]] = i

    def generate_observations_rough(self, conditions):
        # if we have already called for this mjd, no need to repeat.
        if self.last_mjd == conditions.mjd:
            return self.observations

        observations = self._check_list(conditions)
        if observations is None:
            self.observations = []
            self.last_mjd = conditions.mjd
            return self.observations

        n_filter_changes = np.sum(observations[1:]["filter"] == observations[:-1]["filter"])

        # If we want to ensure the observations can be completed
        # before twilight starts
        if self.before_twi_check:
            # Note that if detailers are adding lots of exposures, this
            # calculation has the potential to not be right at all.
            # Also assumes slew time is negligible.
            exptime_needed = np.sum(observations["exptime"]) / 3600.0 / 24.0  # to days
            filter_change_needed = n_filter_changes * self.filter_change_time
            tot_time_needed = exptime_needed + filter_change_needed
            time_before_twi = conditions.sun_n18_rising - conditions.mjd
            # Not enough time, wipe out the observations
            if tot_time_needed > time_before_twi:
                self.observations = []
                self.last_mjd = conditions.mjd
                return self.observations

        self.observations = [self._slice2obs(obs) for obs in observations]
        self.last_mjd = conditions.mjd

        return self.observations



__all__ = ("FilterSwapScheduler", "SimpleFilterSched", "ComCamFilterSched", "FilterSchedUzy")

import numpy as np

from rubin_scheduler.scheduler.utils import IntRounded


class FilterSwapScheduler:
    """A simple way to schedule what filter to load"""

    def __init__(self):
        pass

    def add_observation(self, observation):
        pass

    def __call__(self, conditions):
        """
        Returns
        -------
        list of strings for the filters that should be loaded
        """
        pass


class SimpleFilterSched(FilterSwapScheduler):
    def __init__(self, illum_limit=10.0):
        self.illum_limit_ir = IntRounded(illum_limit)

    def __call__(self, conditions):
        if IntRounded(conditions.moon_phase_sunset) > self.illum_limit_ir:
            result = ["g", "r", "i", "z", "y"]
        else:
            result = ["u", "g", "r", "i", "z"]
        return result


class ComCamFilterSched(FilterSwapScheduler):
    """ComCam can only hold 3 filters at a time.

    Pretend we will cycle from ugr, gri, riz, izy
    depending on lunar phase.

    Parameters
    ----------
    loaded_filter_groups : `tuple` (`tuple`)
        Groups of 3 filters, to be loaded at the same time.
        Multiple groups can be specified, to be swapped between at the
        boundaries of `illum_bins` in lunar phase.
    illum_bins : `np.ndarray`, (N,)
        Lunar illumination boundaries to define when to swap between
        different groups of filters within the `loaded_filter_groups`.
        Lunar illumination ranges from 0 to 100.

    Notes
    -----
    If illum_bins = np.array([0, 50, 100]), then there should be
    two groups of filters to use -- one for use between 0 and 50 percent
    illumination, and another for use between 50 and 100 percent illumination.
    """

    def __init__(
        self,
        loaded_filter_groups=(("u", "g", "r"), ("g", "r", "i"), ("r", "i", "z"), ("i", "z", "y")),
        illum_bins=np.arange(0, 100 + 1, 25),
    ):
        self.loaded_filter_groups = loaded_filter_groups
        self.illum_bins = illum_bins
        if isinstance(self.illum_bins, list):
            self.illum_bins = np.array(illum_bins)
        if len(illum_bins) - 1 > len(loaded_filter_groups):
            raise ValueError("There are illumination bins with an " "undefined loaded_filter_group")

    def __call__(self, conditions):
        moon_at_sunset = conditions.moon_phase_sunset
        try:
            if len(moon_at_sunset) > 0:
                moon_at_sunset = moon_at_sunset[0]
        except TypeError:
            pass
        indx = np.searchsorted(self.illum_bins, moon_at_sunset, side="left")
        indx = np.max([0, indx - 1])
        result = list(self.loaded_filter_groups[indx])
        return result


class FilterSchedUzy(FilterSwapScheduler):
    """
    remove u in bright time. Alternate between removing z and y in
    dark time.

    Note, this might not work properly if we need to restart a bunch.
    So a more robust way of scheduling filter loading might be in order.
    """

    def __init__(self, illum_limit=10.0):
        self.illum_limit_ir = IntRounded(illum_limit)
        self.last_swap = 0

        self.bright_time = ["g", "r", "i", "z", "y"]
        self.dark_times = [["u", "g", "r", "i", "y"], ["u", "g", "r", "i", "z"]]

    def __call__(self, conditions):
        if IntRounded(conditions.moon_phase_sunset) > self.illum_limit_ir:
            result = self.bright_time
        else:
            indx = self.last_swap % 2
            result = self.dark_times[indx]
            if result != conditions.mounted_filters:
                self.last_swap += 1
        return result



__all__ = ("int_binned_stat",)

import numpy as np


def int_binned_stat(ids, values, statistic=np.mean):
    """
    Like scipy.binned_statistic, but for unique integer ids.

    Parameters
    ----------
    ids : array-like of ints
        The integer ID for each value
    values : array-like
        The values to be combined
    statistic : function (np.mean)
        Function to run on the values that have matching ids.

    Returns
    -------
    unique ids, binned values
    """

    uids = np.unique(ids)
    order = np.argsort(ids)

    ordered_ids = ids[order]
    ordered_values = values[order]

    left = np.searchsorted(ordered_ids, uids, side="left")
    right = np.searchsorted(ordered_ids, uids, side="right")

    stat_results = []
    for le, ri in zip(left, right):
        stat_results.append(statistic(ordered_values[le:ri]))

    return uids, np.array(stat_results)
